# Copyright (C) 2014  Andrea De Pasquale <andrea@de-pasquale.name>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging

import conpot.core as conpot_core

logger = logging.getLogger(__name__)


class BaseCommand(object):
    HELP_MESSAGE = ''
    CMD_OUTPUT = ''
    INVALID_PARAMETER = (
        "\r\n"
        "? Invalid parameter.\r\n"
        "Try 'H cmd' for specific help.\r\n"
        " Ie: H !SC\r\n"
    )

    def help(self):
        return self.HELP_MESSAGE

    def run(self, params=None):
        return self.CMD_OUTPUT


class HelpCommand(BaseCommand):
    CMD_OUTPUT = (
        "==============================================================================\r\n"
        "Service Menu\r\n"
        "==============================================================================\r\n"
        "H:   Help [cmd].\r\n"
        "Q:   Close connection.\r\n"
        "!AC: Access control.\r\n"
        "!AS: Alarm Server.\r\n"
        "!GC: Get Config.\r\n"
        "!GV: Software version.\r\n"
        "!SA: Set KAP Server IP and port (*1).\r\n"
        "!SB: Set 2nd KAP Server IP and port.\r\n"
        "!SC: Set Config (*1).\r\n"
        "!SD: Set device name (*1).\r\n"
        "!SH: Set KAP Server lookup (DNS or DHCP)\r\n"
        "!SI: Set IP (enter either valid IP or 0 to force DHCP)(*1).\r\n"
        "!SK: Set KAP watchdog timeout(WDT).\r\n"
        "!SN: Set IP for DNS Name servers to use.\r\n"
        "!SP: Set IP Ports\r\n"
        "!SS: Set Serial Settings.\r\n"
        "!RC: Request connect\r\n"
        "!RR: Request restart (*1).\r\n"
        "!WM: Wink module.\r\n"
        "==============================================================================\r\n"
        "(*1) Forces system restart\r\n"
        "==============================================================================\r\n"
        "Kamstrup (R)\r\n"
    )

    def __init__(self, commands):
        self.commands = commands

    def run(self, params=None):
        if params is None:
            return self.CMD_OUTPUT

        c = params[0:3]
        if c in list(self.commands.keys()):
            return self.commands[c].help()

        return self.INVALID_PARAMETER


class AccessControlCommand(BaseCommand):
    HELP_MESSAGE = (
        "!AC: Access control.\r\n"
        "     Used for simple IP address firewall filtering.\r\n"
        "     If enabled only the listed IP's can assess this module.\r\n"
        "      Format:  !AC [En/Dis [ID IP]]\r\n"
        "      Example: !AC\r\n"
        "               Lists the setup.\r\n"
        "      Example: !AC 0\r\n"
        "               Disables the filter allowing everybody to access.\r\n"
        "      Example: !AC 0 1 192.168.1.211\r\n"
        "               !AC 0 2 10.0.0.1\r\n"
        "               !AC 0 3 195.215.168.45\r\n"
        "               !AC 1\r\n"
        "               Only connections from 192.168.1.211, \r\n"
        "               10.0.0.1 or 195.215.168.45 are possible.\r\n"
    )

    CMD_OUTPUT = (
        "\r\n"
        "{access_control_status} \r\n"
        " [1] {access_control_1}\r\n"
        " [2] {access_control_2}\r\n"
        " [3] {access_control_3}\r\n"
        " [4] {access_control_4}\r\n"
        " [5] {access_control_5}\r\n"
    )

    def set_access_ip(self, number, ip_string):
        databus = conpot_core.get_databus()
        if ip_string.count('.') == 3:
            if any(x in number for x in ['1', '2', '3', '4', '5']):
                acl_number = int(number)
                final_ip = parse_ip(ip_string)
                databus.set_value('access_control_{0}'.format(acl_number), final_ip)

    def run(self, params=None):
        databus = conpot_core.get_databus()
        cmd_output = ''
        if params:
            # return is always OK apparently...
            cmd_output = '\r\nOK\r\n'
            if len(params) == 1 and params == '0':
                databus.set_value("access_control_status", "DISABLED")
            elif len(params) == 1 and params == '1':
                databus.set_value("access_control_status", "ENABLED")
            elif len(params.split(' ')) == 3:
                cmd, acl_number, ip_address = params.split(' ')
                if cmd == '0':
                    self.set_access_ip(acl_number, ip_address)

        return cmd_output + self.CMD_OUTPUT.format(
            access_control_status=databus.get_value("access_control_status"),
            access_control_1=databus.get_value("access_control_1"),
            access_control_2=databus.get_value("access_control_2"),
            access_control_3=databus.get_value("access_control_3"),
            access_control_4=databus.get_value("access_control_4"),
            access_control_5=databus.get_value("access_control_5"))


class AlarmServerCommand(BaseCommand):
    HELP_MESSAGE = (
        "!AS: Alarm Server.\r\n"
        "     Used to set IP and port of server to handle alarm notifications.\r\n"
        "      Format:  !AS [SrvIP [SrvPort]]\r\n"
        "      Example: !AS 195.215.168.45 \r\n"
        "               Alarms are sent to 195.215.168.45.\r\n"
        "      Example: !AS 195.215.168.45 4000\r\n"
        "               Alarms are sent to to port 4000 on 195.215.168.45.\r\n"
        "      Example: !AS 0.0.0.0\r\n"
        "               Alarm reporting is disabled.\r\n"
    )

    CMD_OUTPUT = (
        "\r\n"
        "Alarm server:  {alarm_server_output} "  # no CRLF
    )

    def run(self, params=None):
        databus = conpot_core.get_databus()
        output_prefix = ''
        output_postfix = ''
        if not params:
            if databus.get_value('alarm_server_status') == 'DISABLED':
                output = 'DISABLED'
            else:
                output = '{0}:{1}'.format(databus.get_value('alarm_server_ip'), databus.get_value('alarm_server_port'))
        else:
            output_prefix = '\r\nOK'
            # in this case the command has CRLF... really funky...
            output_postfix = '\r\n'
            databus.set_value('alarm_server_status', 'ENABLED')
            params_split = params.split(' ')
            databus.set_value('alarm_server_ip', parse_ip(params_split[0]))
            # port provided also
            if len(params_split) > 1:
                port = parse_port(params_split[1])
                if port != 0:
                    databus.set_value('alarm_server_port', port)
            output = '{0}:{1}'.format(databus.get_value('alarm_server_ip'), databus.get_value('alarm_server_port'))
        return output_prefix + self.CMD_OUTPUT.format(
            alarm_server_output=output) + output_postfix


class GetConfigCommand(BaseCommand):
    HELP_MESSAGE = (
        "!GC: Get Config.\r\n"
        "     Returns the module configuration.\r\n"
    )

    CMD_OUTPUT = (
        "Device Name         : {device_name}\r\n"
        "Use DHCP            : {use_dhcp}\r\n"
        "IP addr.            : {ip_addr}\r\n"
        "IP Subnet           : {ip_subnet}\r\n"
        "Gateway addr.       : {ip_gateway}\r\n"
        "Service server addr.: {service_server_ip}\r\n"
        "Service server hostname.: {service_server_host}\r\n"
        "DNS Server No. 1: {nameserver_1}\r\n"
        "DNS Server No. 2: {nameserver_2}\r\n"
        "DNS Server No. 3: {nameserver_3}\r\n"
        "MAC addr. (HEX)     : {mac_address}\r\n"
        # TODO: i think was can get these from the other protocol also
        "Channel A device meterno.: {channel_a_meternumber}\r\n"
        "Channel B device meterno.: {channel_b_meternumber}\r\n"
        # TODO: these...
        "Keep alive timer (flash setting): ENABLED 10\r\n"
        "Keep alive timer (current setting): ENABLED 10\r\n"
        "Has the module received acknowledge from the server: {kap_ack_server}\r\n"
        "KAP Server port: {kap_a_server_port}\r\n"
        "KAP Local port: {kap_local_port}\r\n"
        # TODO: This, read from other proto also?
        "Software watchdog: ENABLED 3600\r\n"
    )

    def run(self, params=None):
        databus = conpot_core.get_databus()
        return self.CMD_OUTPUT.format(
            device_name=databus.get_value("device_name"),
            nameserver_1=databus.get_value("nameserver_1"),
            nameserver_2=databus.get_value("nameserver_2"),
            nameserver_3=databus.get_value("nameserver_3"),
            mac_address=databus.get_value("mac_address"),
            use_dhcp=databus.get_value("use_dhcp"),
            ip_addr=databus.get_value("ip_addr"),
            ip_subnet=databus.get_value("ip_subnet"),
            ip_gateway=databus.get_value("ip_gateway"),
            service_server_ip=databus.get_value("kap_a_server_ip"),
            service_server_host=databus.get_value("kap_a_server_hostname"),
            channel_a_meternumber=databus.get_value("channel_a_meternumber"),
            channel_b_meternumber=databus.get_value("channel_b_meternumber"),
            kap_ack_server=databus.get_value("kap_ack_server"),
            kap_a_server_port=databus.get_value("kap_a_server_port"),
            kap_local_port=databus.get_value("kap_local_port"), )


class SoftwareVersionCommand(BaseCommand):
    HELP_MESSAGE = (
        "!GV: Software version.\r\n"
        "     Returns the software revision of the module.\r\n"
    )

    CMD_OUTPUT = (
        "\r\n"
        "Software Version: {software_version}\r\n"
    )

    def run(self, params=None):
        return self.CMD_OUTPUT.format(
            software_version=conpot_core.get_databus().get_value("software_version"))


class SetKap1Command(BaseCommand):
    HELP_MESSAGE = (
        "!SA: Set KAP Server IP and port (*1).\r\n"  # restart is not forced...
        "     Used for setting the IP of the Server to receive KAP-pacakeges.\r\n"
        "     UDP port on server can be provided optionally.\r\n"
        "      Format:  !SA SrvIP [SrvPort]\r\n"
        "      Example: !SA 195215168045 \r\n"
        "               KAP packages are hereafter sent to 195.215.168.45.\r\n"
        "      Example: !SA 195.215.168.45 \r\n"
        "               Same result as \"!SA 195215168045\".\r\n"
        "      Example: !SA 192168001002 61000\r\n"
        "               KAP packages are hereafter sent to 192.168.1.2:61000\r\n"
        "               from module port 8000.\r\n"
        "      Example: !SA 0.0.0.0 \r\n"
        "               Disables KAP.\r\n"
    )

    CMD_OUTPUT = (
        "\r\n"
        "Service server addr.: {kap_a_output}\r\n"
    )

    def run(self, params=None):
        databus = conpot_core.get_databus()
        if params:
            output_prefix = '\r\nOK'
            params_split = params.split(' ')
            databus.set_value('kap_a_server_ip', parse_ip(params_split[0]))
            # TODO: The meter might do a lookup on the ip, and the result of that
            # lookup might be stored in a_server_host...
            databus.set_value('kap_a_server_hostname', '0 - none')
            # port provided also
            if len(params_split) > 1:
                port = parse_port(params_split[1])
                if port != 0:
                    databus.set_value('kap_a_server_port', port)
        else:
            output_prefix = '\r\n'
        output = '{0}:{1}'.format(databus.get_value('kap_a_server_ip'), databus.get_value('kap_a_server_port'))
        return output_prefix + self.CMD_OUTPUT.format(kap_a_output=output)


class SetKap2Command(BaseCommand):
    HELP_MESSAGE = (
        "!SB: Set 2nd KAP Server IP and port.\r\n"
        "     Used for redundancy with two KAP servers.\r\n"
        "     When enabled every second KAP is send to the IP defined by !SB.\r\n"
        "     NB: The KAP interval to each individual server is half of KAPINT\r\n"
        "         defined by !SK.\r\n"
        "     NB: !SA must be enabled (not 0.0.0.0) \r\n"
        "      Format:  !SB SrvIP [SrvPort]\r\n"
        "      Example: !SB 195.215.168.45 \r\n"
        "               KAP packages are hereafter also sent to 195.215.168.45.\r\n"
        "      Example: !SB 0.0.0.0 \r\n"
        "               Disabled.\r\n"
        "      Example: !SB 192.168.1.2 61000\r\n"
        "               KAP packages are hereafter sent to 192.168.1.2:61000\r\n"
        "               from module port 8000.\r\n"
    )

    CMD_OUTPUT_SINGLE = (
        "\r\n"
        "{}\r\n"
        "Service server addr.: {}:{} (from DNS)\r\n"
        "No redundancy."
    )

    CMD_OUTPUT_DOUBLE = (
        "\r\n"
        "{}\r\n"
        "Service server addr.: {}:{} (from DNS)\r\n"
        "and fallback KAP to:  {}:{}\r\n"
    )

    def run(self, params=None):
        databus = conpot_core.get_databus()
        cmd_ok = ""
        if params:
            cmd_ok = "OK"
            params_split = params.split(" ")
            databus.set_value("kap_b_server_ip", parse_ip(params_split[0]))
            if len(params_split) > 1:
                port = parse_port(params_split[1])
                if port != 0:
                    databus.set_value("kap_b_server_port", params_split[1])

        if databus.get_value("kap_b_server_ip") == "0.0.0.0":
            return self.CMD_OUTPUT_SINGLE.format(cmd_ok,
                databus.get_value("kap_a_server_ip"),
                databus.get_value("kap_a_server_port"))
        return self.CMD_OUTPUT_DOUBLE.format(cmd_ok,
            databus.get_value("kap_a_server_ip"),
            databus.get_value("kap_a_server_port"),
            databus.get_value("kap_b_server_ip"),
            databus.get_value("kap_b_server_port"))


class SetConfigCommand(BaseCommand):
    HELP_MESSAGE = (
        "!SC: Set Config (*1).\r\n"
        "     Configures the module.\r\n"
        "      Format:  !SC DHCP IP SUB GW DNS1 DNS2 DNS3 SRV_IP DEVICENAME SRV_DNS\r\n"
        "               DHCP        1 for DHCP, 0 for static IP.\r\n"
        "               IP..        Static IP settings.\r\n"
        "               SRV_IP      IP of remote server (Only if SRV_DNS is 0).\r\n"
        "               DEVICENAME  User label for for individual naming.\r\n"
        "               SRV_DNS     DNS name of remote server (0 to disable DNS lookup)\r\n"
    )

    CMD_OUTPUT = (
        "\r\n"
        "Service server hostname.: {}\r\n"
    )

    def run(self, params=None):
        databus = conpot_core.get_databus()
        if params:
            params_split = params.split(" ")
            if len(params_split) >= 10:

                if params_split[0] == "1":
                    databus.set_value("use_dhcp", "YES")
                else:
                    databus.set_value("use_dhcp", "NO")
                    databus.set_value("ip_addr", parse_ip(params_split[1]))
                    databus.set_value("ip_subnet", parse_ip(params_split[2]))
                    databus.set_value("ip_gateway", parse_ip(params_split[3]))

                databus.set_value("nameserver_1", parse_ip(params_split[4]))
                databus.set_value("nameserver_2", parse_ip(params_split[5]))
                databus.set_value("nameserver_3", parse_ip(params_split[6]))

                if params_split[9] == "0":
                    databus.set_value("kap_a_server_ip", parse_ip(params_split[7]))
                    databus.set_value("kap_a_server_hostname", "0 - none")
                else:
                    databus.set_value("kap_a_server_hostname", params_split[9])
                    # FIXME: server IP should be resolved from the hostname
                    # using nameserver_1, nameserver_2, nameserver_3
                    databus.set_value("kap_a_server_ip", parse_ip(params_split[7]))

                device_name = params_split[8]
                if len(device_name) > 20:
                    device_name = device_name[0:20]
                databus.set_value("device_name", device_name)

        databus.set_value("reboot_signal", 1)


class SetDeviceNameCommand(BaseCommand):
    HELP_MESSAGE = (
        "!SD: Set device name (*1).\r\n"
        "     Option for individual naming of the module (0-20 chars).\r\n"
    )

    def run(self, params=None):
        if params is None:
            params = ""

        if len(params) > 20:
            params = params[0:20]
            output = ""
        else:
            output = "\r\nOK"

        databus = conpot_core.get_databus()
        databus.set_value("device_name", params)
        databus.set_value("reboot_signal", 1)
        return output


class SetLookupCommand(BaseCommand):
    HELP_MESSAGE = (
        "!SH: Set KAP Server lookup (DNS or DHCP)\r\n"
        "     Used for setting the DNS name of the Server to receive KAP-pacakeges.\r\n"
        "     Using the keyword \"DHCP_OPTION:xxx\" makes the module request the IP using DHCP option xxx.\r\n"
        "     The settings are first activated when the module is reset (using !RR).\r\n"
        "      Example: !SH 0 \r\n"
        "               Lookup Disabled.\r\n"
        "               The module will send KAP to the IP listed by !SA. \r\n"
        "      Example: !SH hosting.kamstrup_meter.dk \r\n"
        "               Use DNS lookup.\r\n"
        "               The module will send KAP to the IP listed by !SA until it resolves the DNS,\r\n"
        "               hereafter the KAP will be sent to hosting.kamstrup_meter.dk\r\n"
        "      Example: !SH DHCP_OPTION:129\r\n"
        "               Use DHCP Option.\r\n"
        "               The module will send KAP to the IP provided by DHCP (in this case option 129).\r\n"
        "               The module uses the IP provided by !SA if the DHCP offer dos not include option xxx data.\r\n"
    )

    def run(self, params=None):
        if params is None:
            params = ""

        output = "\r\n"

        databus = conpot_core.get_databus()
        # no, i am not making this up... this is actually how it is implemented on the Kamstrup meter..
        if len(params) == 1:
            databus.set_value('kap_server_lookup', '0 - none')
            output = "\r\nOK" + output
        elif len(params) > 1:
            databus.set_value('kap_server_lookup', params)
            output = "\r\nOK" + output

        output += "Service server hostname.: {0}\r\n"
        return output.format(databus.get_value('kap_server_lookup'))


class SetIPCommand(BaseCommand):
    HELP_MESSAGE = (
        "!SI: Set IP (enter either valid IP or 0 to force DHCP)(*1).\r\n"
        "     Used for changing the module IP.\r\n"
        "     (Use !SC if you need to change subnet/Gateway too).\r\n"
        "     Entering a '0' will enable DHCP.\r\n"
        "      Format:  !SI IP\r\n"
        "      Example: !SI 0\r\n"
        "               The module will reboot and acuire the IP settings using DHCP.\r\n"
        "      Example: !SI 192168001200\r\n"
        "               The module will reboot using static IP addr 192.168.1.200.\r\n"
        "               (SUB, GW and DNS unchanged)\r\n"
        "      Example: !SI 192.168.1.200\r\n"
        "               Same as !SI 192168001200.\r\n"
    )

    CMD_OUTPUT = (
        "\r\n"
        "Use DHCP            : {use_dhcp}\r\n"
        "\r\n"
        "IP addr.            : {ip_addr}\r\n"
    )

    def run(self, params=None):
        databus = conpot_core.get_databus()
        if params:
            ip_addr = parse_ip(params)
            if ip_addr == "0.0.0.0":
                if databus.get_value("use_dhcp") == "NO":
                    databus.set_value("use_dhcp", "YES")
                    databus.set_value("ip_addr",
                        databus.get_value("ip_addr_dhcp"))
                    databus.set_value("ip_gateway",
                        databus.get_value("ip_gateway_dhcp"))
                    databus.set_value("ip_subnet",
                        databus.get_value("ip_subnet_dhcp"))
                    databus.set_value("reboot_signal", 1)
            else:
                databus.set_value("use_dhcp", "NO")
                databus.set_value("ip_addr", ip_addr)
                databus.set_value("reboot_signal", 1)

        return self.CMD_OUTPUT.format(
            use_dhcp=databus.get_value("use_dhcp"),
            ip_addr=databus.get_value("ip_addr"))


class SetWatchdogCommand(BaseCommand):
    HELP_MESSAGE = (
        "!SK: Set KAP watchdog timeout(WDT).\r\n"
        "     Used for setting KeepAlive watchdog timing.\r\n"
        "      Format:  !SK [WDT] [MISSING] [KAPINT]\r\n"
        "      Example: !SK\r\n"
        "      Example: !SK 3600\r\n"
        "      Example: !SK 3600 60 10\r\n"
        "               WDT     The module reboots after WDT?KAPINT seconds\r\n"
        "                       without an ACK from the server.\r\n"
        "                       0 = disable WDT.\r\n"
        "               MISSING After MISSING?KAPINT seconds without an ACK,\r\n"
        "                       the Err LED starts blinking.\r\n"
        "                       (Used for indication of missing link to the server)\r\n"
        "               KAPINT  Interval in seconds for how often KeepAlivePackages\r\n"
        "                       are send to the KAP server.\r\n"
        "     The WDT and MISSING timeout counts are both reset by an ACK from the server.                       \r\n"
    )

    CMD_OUTPUT = (
        "Software watchdog: {0}\r\n"
        "KAP Missing warning: {1}\r\n"
        "Keep alive timer (flash setting): {2}\r\n"
    )

    def run(self, params=None):
        output = "\r\n"
        databus = conpot_core.get_databus()

        if params is not None:
            params_split = params.split(" ")

            if len(params_split) > 0:
                # meh, actually the real value is non-existing. If you supply a larger value the smart meter
                # just overwrite memory and starts writing to the next memory location - yep, you heard it here first!
                watchdog_value = str(try_parse_uint(params_split[0], min_value=5, max_value=4294967295))
                databus.set_value('software_watchdog', watchdog_value)
                if len(params_split) > 1:
                    kap_missing = str(try_parse_uint(params_split[1], min_value=0, max_value=4294967295))
                    databus.set_value('kap_missing_warning', kap_missing)
                if len(params_split) > 2:
                    keep_alive_timer = str(try_parse_uint(params_split[2], min_value=0, max_value=4294967295))
                    databus.set_value('keep_alive_timer', keep_alive_timer)
                output = "\r\nOK" + output

        return_values = [databus.get_value('software_watchdog'),
                         databus.get_value('kap_missing_warning'),
                         databus.get_value('keep_alive_timer')]

        for i in range(0, len(return_values)):
            if return_values[i] == '0':
                return_values[i] = 'DISABLED {0}'.format(return_values[i])
            else:
                return_values[i] = 'ENABLED {0}'.format(return_values[i])

        output += SetWatchdogCommand.CMD_OUTPUT.format(return_values[0],
                                                       return_values[1],
                                                       return_values[2])

        return output.format(databus.get_value('kap_server_lookup'))


class SetNameserverCommand(BaseCommand):
    HELP_MESSAGE = (
        "!SN: Set IP for DNS Name servers to use.\r\n"
        "      Format:  !SN DNS1 DNS2 DNS3\r\n"
        "      Example: !SN 192168001200 192168001201 000000000000\r\n"
        "      Example: !SN 172.16.0.83 172.16.0.84 0.0.0.0\r\n"
    )

    def run(self, params=None):
        if params is None:
            return self.INVALID_PARAMETER

        nameservers = params.split(" ")
        if len(nameservers) != 3:
            return self.INVALID_PARAMETER

        databus = conpot_core.get_databus()
        databus.set_value("nameserver_1", parse_ip(nameservers[0]))
        databus.set_value("nameserver_2", parse_ip(nameservers[1]))
        databus.set_value("nameserver_3", parse_ip(nameservers[2]))
        return "\r\nOK"


class SetPortsCommand(BaseCommand):
    HELP_MESSAGE = (
        "!SP: Set IP Ports\r\n"
        "      Format:  !SP [KAP CHA CHB CFG]\r\n"
        "      Example: !SP 333\r\n"
        "               KAP packages are hereafter sent to port 333 on the server.\r\n"
        "      Example: !SP 50 1025 1026 50100\r\n"
        "               KAP packages are sent to port 50.\r\n"
        "               Direct connections to UART channel A is on port 1025, B on 1026.\r\n"
        "               Config connection on port 50100.\r\n"
        "               (default values)\r\n"
        "      Example: !SP 0 0 80\r\n"
        "               UART channel B is on port 80 (KAP and ChA is ingored - unchanged).\r\n"
    )

    CMD_OUTPUT = (
        "\r\n"
        "{}\r\n"
        "KAP on server: {}\r\n"
        "ChA on module: {}\r\n"
        "ChB on module: {}\r\n"
        "Cfg on module: {}\r\n"
    )

    def run(self, params=None):
        databus = conpot_core.get_databus()
        cmd_ok = ""
        if params:
            params_split = params.split(" ")
            cmd_ok = "OK"

            kap_port = parse_port(params_split[0])
            if kap_port != 0:
                databus.set_value("kap_a_server_port", kap_port)

            if len(params_split) > 1:
                cha_port = parse_port(params_split[1])
                if cha_port != 0:
                    databus.set_value("channel_a_port", cha_port)

            if len(params_split) > 2:
                chb_port = parse_port(params_split[2])
                if chb_port != 0:
                    databus.set_value("channel_b_port", chb_port)

            # FIXME: how do we change the port we are connected to?
            #if len(params_split) > 3:
                #cfg_port = parse_port(params_split[3])
                #if cfg_port != 0:
                    #databus.set_value("", cfg_port)

        return self.CMD_OUTPUT.format(cmd_ok,
            databus.get_value("kap_a_server_port"),
            databus.get_value("channel_a_port"),
            databus.get_value("channel_b_port"),
            50100)  # FIXME: see above


class SetSerialCommand(BaseCommand):
    HELP_MESSAGE = (
        "!SS: Set Serial Settings.\r\n"
        "     Used for setting the serial interface for channel A or B.\r\n"
        "      Format:  !SS [Channel Baud,DataBits,Parity,StopBits[,Ctrl]]\r\n"
        "      Example: !SS A Auto\r\n"
        "      Example: !SS A 9600,8,N,2\r\n"
        "      Example: !SS B 115200,8,E,1\r\n"
        "      Example: !SS B 115200,8,E,1,I\r\n"
        "      Example: !SS B 115200,8,E,1,L\r\n"
        "     The ctrl flag can be 'C'(check), 'I' (ignore framing errors) or 'L' (Link, ChB only).\r\n"
        "     Chanel A supports auto mode (Also enables load profile logger in old E-Meters).\r\n"
    )

    def run(self, params=None):
        databus = conpot_core.get_databus()
        invalid_message = "\r\nInvalid data!\r\n\r\n"
        if params:
            params_split = params.split(" ")
            if len(params_split) == 2:
                output = "\r\nOK\r\n"
                if params_split[0] == "A":
                    databus.set_value('serial_settings_a', params_split[1])
                elif params_split[0] == "B":
                    databus.set_value('serial_settings_b', params_split[1])
                else:
                    return invalid_message
            else:
                return invalid_message
        else:
            output = "\r\n" \
                     "UART A setup : {0}\r\n" \
                     "UART B setup : {1}\r\n"

        return output.format(databus.get_value('serial_settings_a'),
                             databus.get_value('serial_settings_b'))


class RequestConnectCommand(BaseCommand):
    HELP_MESSAGE = (
        "!RC: Request connect\r\n"
        "     Makes the module crate a ChA or ChB socket to a remote server.\r\n"
        "      Format:  !RC Action [IP [Port]]\r\n"
        "      Example: !RC A 195.215.168.45 200\r\n"
        "      Example: !RC B 195.215.168.45 201\r\n"
        "      Example: !RC D\r\n"
        "               Disconnects both A and B if open.\r\n"
        "      Example: !RC\r\n"
        "               Status only.\r\n"
    )

    def run(self, params):
        databus = conpot_core.get_databus()
        # TODO: Further investigations needed... How does this remote socket work? How should copot react?
        output = "Status: 0100\r\n"
        if params:
            params_split = params.split(" ")
            output = "\r\nOK\r\n" + output
            if len(params_split) == 1 and params_split[0] == 'D':
                pass
            elif len(params_split) == 2:
                channel, value = params_split
                if channel == "A":
                    # TODO: figure out how these are parsed when meter is online again
                    databus.set_value('channel_a_connect_socket', value)
                elif channel == "B":
                    databus.set_value('channel_b_connect_socket', value)
                else:
                    return self.INVALID_PARAMETER
            else:
                return self.INVALID_PARAMETER
        else:
            output = "\r\n" + output

        return output


class RequestRestartCommand(BaseCommand):
    HELP_MESSAGE = (
        "!RR: Request restart (*1).\r\n"
    )

    def run(self, params=None):
        conpot_core.get_databus().set_value("reboot_signal", 1)
        return


class WinkModuleCommand(BaseCommand):
    HELP_MESSAGE = (
        "!WM: Wink module.\r\n"
        "     Causes the WINK LED on the module to blink for physical identification.\r\n"
    )

    # no other output
    CMD_OUTPUT = (
        "\r\n"
        "\r\n"
        "OK\r\n"
    )


def parse_ip(ip_string):
    default = "0.0.0.0"
    if "." in ip_string:
        octets = ip_string.split(".")
    else:
        octets = [int(ip_string[i:i + 3]) for i in range(0, len(ip_string), 3)]

    if len(octets) is not 4:
        return default
    for octet in octets:
        if int(octet) < 0 or int(octet) > 255:
            return default
    return ".".join(list(map(str, octets)))


def parse_port(port_string):
    try:
        port = int(port_string)
        if 0 < port < 65536:
            return port
        return 0
    except ValueError:
        return 0


def try_parse_uint(uint_string, min_value=0, max_value=254):
    try:
        value = int(uint_string)
        if value < min_value or value > max_value:
            value = 0
    except ValueError:
        value = '0'
    return value



"""
Data related to Kamstrup Management Protocol
"""

RESPONSES = {
    'H': {
        'H': b'==============================================================================\r\nService Menu\r\n==============================================================================\r\nH:   Help [cmd].\r\nQ:   Close connection.\r\n!AC: Access control.\r\n!AS: Alarm Server.\r\n!GC: Get Config.\r\n!GV: Software version.\r\n!SA: Set KAP Server IP and port (*1).\r\n!SB: Set 2nd KAP Server IP and port.\r\n!SC: Set Config (*1).\r\n!SD: Set device name (*1).\r\n!SH: Set KAP Server lookup (DNS or DHCP)\r\n!SI: Set IP (enter either valid IP or 0 to force DHCP)(*1).\r\n!SK: Set KAP watchdog timeout(WDT).\r\n!SN: Set IP for DNS Name servers to use.\r\n!SP: Set IP Ports\r\n!SS: Set Serial Settings.\r\n!RC: Request connect\r\n!RR: Request restart (*1).\r\n!WM: Wink module.\r\n==============================================================================\r\n(*1) Forces system restart\r\n==============================================================================\r\nKamstrup (R)\r\n',
        '!AC': b"!AC: Access control.\r\n     Used for simple IP address firewall filtering.\r\n     If enabled only the listed IP's can assess this module.\r\n      Format:  !AC [En/Dis [ID IP]]\r\n      Example: !AC\r\n               Lists the setup.\r\n      Example: !AC 0\r\n               Disables the filter allowing everybody to access.\r\n      Example: !AC 0 1 192.168.1.211\r\n               !AC 0 2 10.0.0.1\r\n               !AC 0 3 195.215.168.45\r\n               !AC 1\r\n               Only connections from 192.168.1.211, \r\n               10.0.0.1 or 195.215.168.45 are possible.\r\n",
        '!AS': b'!AS: Alarm Server.\r\n     Used to set IP and port of server to handle alarm notifications.\r\n      Format:  !AS [SrvIP [SrvPort]]\r\n      Example: !AS 195.215.168.45 \r\n               Alarms are sent to 195.215.168.45.\r\n      Example: !AS 195.215.168.45 4000\r\n               Alarms are sent to to port 4000 on 195.215.168.45.\r\n      Example: !AS 0.0.0.0\r\n               Alarm reporting is disabled.\r\n',
        '!GC': b'!GC: Get Config.\r\n     Returns the module configuration.\r\n',
        '!GV': b'!GV: Software version.\r\n     Returns the software revision of the module.\r\n',
        '!SA': b'!SA: Set KAP Server IP and port (*1).\r\n     Used for setting the IP of the Server to receive KAP-pacakeges.\r\n     UDP port on server can be provided optionally.\r\n      Format:  !SA SrvIP [SrvPort]\r\n      Example: !SA 195215168045 \r\n               KAP packages are hereafter sent to 195.215.168.45.\r\n      Example: !SA 195.215.168.45 \r\n               Same result as "!SA 195215168045".\r\n      Example: !SA 192168001002 61000\r\n               KAP packages are hereafter sent to 192.168.1.2:61000\r\n               from module port 8000.\r\n      Example: !SA 0.0.0.0 \r\n               Disables KAP.\r\n',
        '!SB': b'!SB: Set 2nd KAP Server IP and port.\r\n     Used for redundancy with two KAP servers.\r\n     When enabled every second KAP is send to the IP defined by !SB.\r\n     NB: The KAP interval to each individual server is half of KAPINT\r\n         defined by !SK.\r\n     NB: !SA must be enabled (not 0.0.0.0) \r\n      Format:  !SB SrvIP [SrvPort]\r\n      Example: !SB 195.215.168.45 \r\n               KAP packages are hereafter also sent to 195.215.168.45.\r\n      Example: !SB 0.0.0.0 \r\n               Disabled.\r\n      Example: !SB 192.168.1.2 61000\r\n               KAP packages are hereafter sent to 192.168.1.2:61000\r\n               from module port 8000.\r\n',
        '!SC': b'!SC: Set Config (*1).\r\n     Configures the module.\r\n      Format:  !SC DHCP IP SUB GW DNS1 DNS2 DNS3 SRV_IP DEVICENAME SRV_DNS\r\n               DHCP        1 for DHCP, 0 for static IP.\r\n               IP..        Static IP settings.\r\n               SRV_IP      IP of remote server (Only if SRV_DNS is 0).\r\n               DEVICENAME  User label for for individual naming.\r\n               SRV_DNS     DNS name of remote server (0 to disable DNS lookup)\r\n',
        '!SD': b'!SD: Set device name (*1).\r\n     Option for individual naming of the module (0-20 chars).\r\n',
        '!SH': b'!SH: Set KAP Server lookup (DNS or DHCP)\r\n     Used for setting the DNS name of the Server to receive KAP-pacakeges.\r\n     Using the keyword "DHCP_OPTION:xxx" makes the module request the IP using DHCP option xxx.\r\n     The settings are first activated when the module is reset (using !RR).\r\n      Example: !SH 0 \r\n               Lookup Disabled.\r\n               The module will send KAP to the IP listed by !SA. \r\n      Example: !SH hosting.kamstrup_meter.dk \r\n               Use DNS lookup.\r\n               The module will send KAP to the IP listed by !SA until it resolves the DNS,\r\n               hereafter the KAP will be sent to hosting.kamstrup_meter.dk\r\n      Example: !SH DHCP_OPTION:129\r\n               Use DHCP Option.\r\n               The module will send KAP to the IP provided by DHCP (in this case option 129).\r\n               The module uses the IP provided by !SA if the DHCP offer dos not include option xxx data.\r\n',
        '!SI': b"!SI: Set IP (enter either valid IP or 0 to force DHCP)(*1).\r\n     Used for changing the module IP.\r\n     (Use !SC if you need to change subnet/Gateway too).\r\n     Entering a '0' will enable DHCP.\r\n      Format:  !SI IP\r\n      Example: !SI 0\r\n               The module will reboot and acuire the IP settings using DHCP.\r\n      Example: !SI 192168001200\r\n               The module will reboot using static IP addr 192.168.1.200.\r\n               (SUB, GW and DNS unchanged)\r\n      Example: !SI 192.168.1.200\r\n               Same as !SI 192168001200.\r\n",
        '!SK': b'!SK: Set KAP watchdog timeout(WDT).\r\n     Used for setting KeepAlive watchdog timing.\r\n      Format:  !SK [WDT] [MISSING] [KAPINT]\r\n      Example: !SK\r\n      Example: !SK 3600\r\n      Example: !SK 3600 60 10\r\n               WDT     The module reboots after WDT?KAPINT seconds\r\n                       without an ACK from the server.\r\n                       0 = disable WDT.\r\n               MISSING After MISSING?KAPINT seconds without an ACK,\r\n                       the Err LED starts blinking.\r\n                       (Used for indication of missing link to the server)\r\n               KAPINT  Interval in seconds for how often KeepAlivePackages\r\n                       are send to the KAP server.\r\n     The WDT and MISSING timeout counts are both reset by an ACK from the server.                       \r\n',
        '!SN': b'!SN: Set IP for DNS Name servers to use.\r\n      Format:  !SN DNS1 DNS2 DNS3\r\n      Example: !SN 192168001200 192168001201 000000000000\r\n      Example: !SN 172.16.0.83 172.16.0.84 0.0.0.0\r\n',
        '!SP': b'!SP: Set IP Ports\r\n      Format:  !SP [KAP CHA CHB CFG]\r\n      Example: !SP 333\r\n               KAP packages are hereafter sent to port 333 on the server.\r\n      Example: !SP 50 1025 1026 50100\r\n               KAP packages are sent to port 50.\r\n               Direct connections to UART channel A is on port 1025, B on 1026.\r\n               Config connection on port 50100.\r\n               (default values)\r\n      Example: !SP 0 0 80\r\n               UART channel B is on port 80 (KAP and ChA is ingored - unchanged).\r\n',
        '!SS': b"!SS: Set Serial Settings.\r\n     Used for setting the serial interface for channel A or B.\r\n      Format:  !SS [Channel Baud,DataBits,Parity,StopBits[,Ctrl]]\r\n      Example: !SS A Auto\r\n      Example: !SS A 9600,8,N,2\r\n      Example: !SS B 115200,8,E,1\r\n      Example: !SS B 115200,8,E,1,I\r\n      Example: !SS B 115200,8,E,1,L\r\n     The ctrl flag can be 'C'(check), 'I' (ignore framing errors) or 'L' (Link, ChB only).\r\n     Chanel A supports auto mode (Also enables load profile logger in old E-Meters).\r\n",
        '!RC': b'!RC: Request connect\r\n     Makes the module crate a ChA or ChB socket to a remote server.\r\n      Format:  !RC Action [IP [Port]]\r\n      Example: !RC A 195.215.168.45 200\r\n      Example: !RC B 195.215.168.45 201\r\n      Example: !RC D\r\n               Disconnects both A and B if open.\r\n      Example: !RC\r\n               Status only.\r\n'
    },
    '!AC': b'\r\nOK\r\n\r\nDISABLED \r\n [1] 0.0.0.0\r\n [2] 192.168.1.211\r\n [3] 0.0.0.0\r\n [4] 0.0.0.0\r\n [5] 0.0.0.0\r\n',
    '!AS': b'\r\nOK\r\n\r\nDISABLED \r\n [1] 0.0.0.0\r\n [2] 0.0.0.0\r\n [3] 0.0.0.0\r\n [4] 0.0.0.0\r\n [5] 0.0.0.0\r\n',
    '!GC': b'Device Name         : \r\nUse DHCP            : YES\r\nIP addr.            : 192.168.1.210\r\nIP Subnet           : 255.255.255.0\r\nGateway addr.       : 192.168.1.1\r\nService server addr.: 202.202.202.1\r\nService server hostname.: kapserver.evilpowerprovider.org\r\nDNS Server No. 1: 0.0.0.0\r\nDNS Server No. 2: 0.0.0.0\r\nDNS Server No. 3: 0.0.0.0\r\nMAC addr. (HEX)     : 00:13:EA:00:00:00\r\nChannel A device meterno.: A1 06 A1 02 B7 34 12 00 00 03\r\nChannel B device meterno.: A1 06 A1 02 B7 34 12 00 00 03\r\nKeep alive timer (flash setting): ENABLED 10\r\nKeep alive timer (current setting): ENABLED 10\r\nHas the module received acknowledge from the server: NO\r\nKAP Server port: 50\r\nKAP Local port: 800\r\nSoftware watchdog: ENABLED 3600\r\n',
    '!GV': b'\r\nSoftware Version: 5.5 (E5)\r\n',
    '!SA': b'\r\nOK\r\nService server addr.: 192.168.1.2:61000\r\n',
    '!SB': b'\r\nOK\r\nService server addr.: 202.202.202.1:50 (from DNS)\r\nand fallback KAP to:  192.168.1.2:61000\r\n',
    '!SC': b'!SC: Set Config (*1).\r\n     Configures the module.\r\n      Format:  !SC DHCP IP SUB GW DNS1 DNS2 DNS3 SRV_IP DEVICENAME SRV_DNS\r\n               DHCP        1 for DHCP, 0 for static IP.\r\n               IP..        Static IP settings.\r\n               SRV_IP      IP of remote server (Only if SRV_DNS is 0).\r\n               DEVICENAME  User label for for individual naming.\r\n               SRV_DNS     DNS name of remote server (0 to disable DNS lookup)\r\n',
    '!SD': b'\r\nOK',
    '!SH': b'\r\nOK\r\nService server hostname.: hosting.kamstrup_meter.dk\r\n',
    '!SI': b'\r\nUse DHCP            : NO\r\n\r\nIP addr.            : 192.168.1.200\r\n',
    '!SK': b'\r\nOK\r\nSoftware watchdog: ENABLED 3600\r\nKAP Missing warning: ENABLED 60\r\nKeep alive timer (flash setting): ENABLED 10\r\n',
    '!SN': b'\r\nOK',
    '!SP': b'\r\nOK\r\nKAP on server: 50\r\nChA on module: 1025\r\nChB on module: 1026\r\nCfg on module: 50100\r\n',
    '!SS': b'\r\nOK\r\n',
    '!RC': b'\r\nOK\r\nStatus: 0100\r\n'
}


# Copyright (C) 2014  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import unittest
from conpot.protocols.kamstrup.meter_protocol.decoder_382 import Decoder382


class TestKamstrupDecoder(unittest.TestCase):
    # TODO: Rename functions when i figure out the actual meaning of the requests / responses
    def test_request_one(self):
        request = [chr(0x80), chr(0x3f), chr(0x10), chr(0x01), chr(0x04), chr(0x1e), chr(0x7a), chr(0xbb), chr(0x0d)]
        decoder = Decoder382()
        result = decoder.decode_in(request)
        self.assertEqual(result, 'Request for 1 register(s): 1054 (Voltage p1) [0x3f]')

    def test_invalid_crc(self):
        invalid_sequences = [
            [chr(0x80), chr(0x3f), chr(0x10), chr(0x02), chr(0x00), chr(0x01), chr(0x55), chr(0xa1), chr(0x0d)],
            [chr(0x80), chr(0x3f), chr(0x10), chr(0x01), chr(0x00), chr(0x02), chr(0x65), chr(0xcf), chr(0x0d)]]

        for seq in invalid_sequences:
            decoder = Decoder382()
            result = decoder.decode_in(seq)
            self.assertEqual(result, 'Request discarded due to invalid CRC.',
                             'Invalid CRC {0} tested valid'.format(seq))

            # def test_request_two(self):
            #     request = "803f1001000265c20d".encode('hex-codec')
            #     decoder = Decoder()
            #     result = decoder.decode_in(request)
            #
            # def test_response_one(self):
            #     response = "403f1000010204000000008be1900d".encode('hex-codec')
            #     decoder = Decoder()
            #     result = decoder.decode_in(response)
            #
            # def test_response_two(self):
            #     response = "403f10000202040000000000091bf90d".encode('hex-codec')
            #     decoder = Decoder()
            #     result = decoder.decode_in(response)


# Copyright (C) 2018  Abhinav Saxena <xandfury@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import gevent
from gevent import monkey; gevent.monkey.patch_all()
import unittest
import os
from datetime import datetime
import conpot
from freezegun import freeze_time
from conpot.protocols.ftp.ftp_utils import ftp_commands
import conpot.core as conpot_core
from conpot.helpers import sanitize_file_name
from conpot.protocols.ftp.ftp_server import FTPServer
import ftplib      # Use ftplib's client for more authentic testing


class TestFTPServer(unittest.TestCase):

    """
        All tests are executed in a similar way. We run a valid/invalid FTP request/command and check for valid
        response. Testing is done by sending/receiving files in data channel related commands.
        Implementation Note: There are no explicit tests for active/passive mode. These are covered in list and nlst
        tests
    """

    def setUp(self):
        # Initialize the file system
        conpot_core.initialize_vfs()
        # get the current directory
        self.dir_name = os.path.dirname(conpot.__file__)
        self.ftp_server = FTPServer(self.dir_name + '/templates/default/ftp/ftp.xml', None, None)
        self.server_greenlet = gevent.spawn(self.ftp_server.start, '127.0.0.1', 0)
        self.client = ftplib.FTP()
        gevent.sleep(1)

    def tearDown(self):
        if self.client:
            try:
                self.client.close()
            except ftplib.all_errors:
                pass
        self.ftp_server.stop()
        self.server_greenlet.kill()

    def refresh_client(self):
        """
        Disconnect and reconnect a client
        """
        if self.client:
            self.client.quit()
            del self.client
        self.client = ftplib.FTP()
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)

    def test_auth(self):
        """Test for user, pass and quit commands."""
        # test with anonymous
        self.assertEqual(self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port),
                         '200 FTP server ready.')
        self.assertIn('Technodrome - Mouser Factory.', self.client.login())
        self.refresh_client()
        # test with registered user nobody:nobody
        self.assertIn('Technodrome - Mouser Factory.', self.client.login(user='nobody', passwd='nobody'))
        # testing with incorrect password
        # testing with incorrect username
        # try to access a command that requires auth with being authenticated.
        self.assertEqual(self.client.quit(), '221 Bye.')

    def test_help(self):
        # TODO: test help before login and after login.
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        cmds = self.ftp_server.handler.config.enabled_commands
        [cmds.remove(i) for i in ('SITE HELP', 'SITE CHMOD') if i in cmds]
        help_text = self.client.sendcmd('help')
        self.assertTrue(all([True if i in help_text else False for i in cmds]))
        # test command specific help
        cmds_help = {ftp_commands[k]['help'] for k in cmds}
        self.assertTrue(all([True for i in cmds if self.client.sendcmd('help {}'.format(i)) and i != 'HELP' in
                             cmds_help]))
        # test unrecognized command
        self.assertRaisesRegex(ftplib.error_perm, '501 Unrecognized command.', self.client.sendcmd, 'help ABCD')

    def test_noop(self):
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        self.assertEqual(self.client.sendcmd('noop'), '200 I successfully done nothin\'.')

    def test_stru(self):
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        self.assertEqual(self.client.sendcmd('stru F'), '200 File transfer structure set to: F.')
        self.assertRaisesRegex(ftplib.error_perm, '504 Unimplemented STRU type.', self.client.sendcmd, 'stru P')
        self.assertRaisesRegex(ftplib.error_perm, '504 Unimplemented STRU type.', self.client.sendcmd, 'stru R')
        self.assertRaisesRegex(ftplib.error_perm, '501 Unrecognized STRU type.', self.client.sendcmd,
                               'stru invalid_stru_cmd')

    def test_allo(self):
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        self.assertEqual(self.client.sendcmd('allo 250'), '202 No storage allocation necessary.')

    def test_syst(self):
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        self.assertEqual(self.client.sendcmd('syst'), '215 UNIX Type: L8')

    def test_mode(self):
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        self.assertRaisesRegex(ftplib.error_perm, '501 Syntax error: command needs an argument', self.client.sendcmd,
                               'mode')
        self.assertEqual(self.client.sendcmd('mode S'), '200 Transfer mode set to: S')
        self.assertRaisesRegex(ftplib.error_perm, '504 Unimplemented MODE type.', self.client.sendcmd, 'mode B')

    def test_site(self):
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        self.assertRaisesRegex(ftplib.error_perm, '501 Syntax error: command needs an argument', self.client.sendcmd,
                               'site')

    def test_site_help(self):
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        self.assertIn('Help SITE command successful.', self.client.sendcmd('site help'))
        self.assertIn('HELP', self.client.sendcmd('site help'))
        self.assertIn('CHMOD', self.client.sendcmd('site help'))

    def test_type(self):
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        self.assertEqual(self.client.sendcmd('type I'), '200 Type set to: Binary.')
        self.assertEqual(self.client.sendcmd('type L8'), '200 Type set to: Binary.')
        self.assertEqual(self.client.sendcmd('type A'), '200 Type set to: ASCII.')
        self.assertEqual(self.client.sendcmd('type L7'), '200 Type set to: ASCII.')
        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'type 234')

    def test_size(self):
        # TODO: test for a user who does not has permissions for size
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        self.assertRaisesRegex(ftplib.error_perm, '550 SIZE not allowed in ASCII mode.', self.client.sendcmd,
                               'size ftp_data.txt')
        # change to mode to binary
        _ = self.client.sendcmd('type I')
        self.assertEqual(self.client.sendcmd('size ftp_data.txt'), '213 49')
        self.assertRaisesRegex(ftplib.error_perm, '550 is not retrievable.', self.client.sendcmd,
                               'size file_does_not_exist')

    def test_pwd(self):
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        self.assertEqual(self.client.sendcmd('pwd'), '257 "/" is the current directory.')

    def test_mkd(self):
        # TODO: test for a user who does not has permissions to make directory
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        self.assertEqual(self.client.sendcmd('mkd testing'), '257 "/testing" directory created.')
        self.assertRaisesRegex(ftplib.error_perm,
                               "550 'mkd /../../testing/testing' points to a path which is "
                               "outside the user's root directory.", self.client.sendcmd, 'mkd /../../testing/testing')
        _ = self.client.sendcmd('mkd testing/testing')
        self.assertEqual(self.client.sendcmd('mkd testing/testing/../demo'),
                         '257 "/testing/demo" directory created.')
        _vfs, _ = conpot_core.get_vfs('ftp')
        _vfs.removedir('testing/testing')
        _vfs.removedir('testing/demo')
        _vfs.removedir('testing')

    def test_cwd(self):
        #  TODO: test for a user who does not has permissions to change directory
        _vfs, _ = conpot_core.get_vfs('ftp')
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        # create a directory to cwd to.
        _vfs.makedir('testing')
        self.assertEqual(self.client.sendcmd('cwd testing'), '250 "/testing" is the current directory.')
        # check consistency with pwd
        self.assertEqual(self.client.sendcmd('pwd'), '257 "/testing" is the current directory.')
        # test for cdup.
        self.assertEqual(self.client.sendcmd('cdup'), '250 "/" is the current directory.')
        # make sure that user does not go - out of the root path.
        self.assertRaisesRegex(ftplib.error_perm, "550 'cwd ../' points to a path which is outside the user's "
                                                  "root directory.", self.client.sendcmd, 'cwd ../')
        _vfs.removedir('testing')

    def test_rmd(self):
        _vfs, _ = conpot_core.get_vfs('ftp')
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        # let us create a temp dir for deleting
        _vfs.makedir('tmp')
        self.assertEqual(self.client.sendcmd('rmd tmp'), '250 Directory removed.')
        self.assertRaisesRegex(ftplib.error_perm, '550 Remove directory operation failed.', self.client.sendcmd,
                               'rmd tmp')
        # TODO: Test with user that has no or little permissions.
        # test for a user trying to delete '/'
        self.assertRaisesRegex(ftplib.error_perm, '550 Can\'t remove root directory.', self.client.sendcmd,
                               'rmd /')
        self.assertRaisesRegex(ftplib.error_perm,
                               "550 'rmd ../../' points to a path which is outside the user's root directory.",
                               self.client.sendcmd, 'rmd ../../')

    @freeze_time('2018-07-15 17:51:17')
    def test_mdtm(self):
        # TODO : test for user that does not have permissions for mdtm
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        _vfs, _ = conpot_core.get_vfs('ftp')
        _vfs.settimes('ftp_data.txt', accessed=datetime.now(), modified=datetime.now())
        # test for a file that already exists
        self.assertEqual(self.client.sendcmd('mdtm ftp_data.txt'), '213 20180715175117')
        self.assertRaisesRegex(ftplib.error_perm, "550 /this_file_does_not_exist.txt is not retrievable",
                               self.client.sendcmd, 'mdtm this_file_does_not_exist.txt')

    def test_dele(self):
        # TODO: check for a user who does not have permissions to delete a file!
        _vfs, _ = conpot_core.get_vfs('ftp')
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        # let us create a temp file just for deleting.
        with _vfs.open('/temp_file', mode='w') as _tmp:
            _tmp.write('This is just a temp file for testing rm')
        # delete that file
        self.assertEqual(self.client.sendcmd('dele temp_file'), '250 File removed.')
        # check for errors
        self.assertRaisesRegex(ftplib.error_perm, '550 Failed to delete file.', self.client.sendcmd, 'dele temp_file')

    def test_file_rename(self):
        # TODO: check for a user who does not have permissions to rename a file!
        _vfs, _ = conpot_core.get_vfs('ftp')
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        # First we would do everything for a valid file and all valid params
        # check with invalid rnfr params
        self.assertRaisesRegex(ftplib.error_perm, "550 Can't rename home directory.", self.client.sendcmd, 'rnfr /')
        self.assertRaisesRegex(ftplib.error_perm, '550 No such file or directory.', self.client.sendcmd,
                               'rnfr file_DNE')
        self.assertRaisesRegex(ftplib.error_perm, "503 Bad sequence of commands: use RNFR first.", self.client.sendcmd,
                               'rnto /random_path')
        # create a custom file to play with.
        try:
            # do a rnfr to rename file ftp_data.txt
            with _vfs.open('/test_rename_file.txt', mode='w') as _test:
                _test.write('This is just a test file for rename testing of FTP server')
            self.assertEqual(self.client.sendcmd('rnfr test_rename_file.txt'), '350 Ready for destination name.')
            self.assertEqual(self.client.sendcmd('rnto new_data.txt'), '250 Renaming ok.')
            # try for a case that would fail --
            # fixme: tests fail after trying to rename files once they have been renamed.
            # self.assertEqual(self.client.sendcmd('rnfr new_data.txt'), '350 Ready for destination name.')
            # self.assertRaisesRegex(ftplib.error_perm, '501 can\'t decode command.', self.client.sendcmd,
            #                        'rnto Very / Unsafe / file\nname hähä \n\r .txt')
        finally:
            _vfs.remove('new_data.txt')

    def test_site_chmod(self):
        # TODO: check for a user who does not have permissions to do chmod!
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        # change permissions
        self.client.sendcmd('site chmod 644 ftp_data.txt')
        _vfs, _ = conpot_core.get_vfs('ftp')
        self.assertEqual(_vfs.get_permissions('ftp_data.txt'), 'rw-r--r--')

    def test_stat(self):
        # TODO: check for a user who does not have permissions to do stat!
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        # do stat without args
        self.assertIn('Logged in as: nobody\n TYPE: ASCII; STRUcture: File; MODE: Stream\n',
                      self.client.sendcmd('stat'))
        self.assertIn('ftp_data.txt', self.client.sendcmd('stat /'))

    # ------ Data channel related. -----

    @freeze_time('2018-07-15 17:51:17')
    def test_list(self):
        # TODO: check for a user who does not have permissions to do list!
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        _vfs, _ = conpot_core.get_vfs('ftp')
        _vfs.settimes('ftp_data.txt', accessed=datetime.now(), modified=datetime.now())
        # Do a list of directory for passive mode
        _pasv_list = list()
        self.client.retrlines('LIST', _pasv_list.append)
        # note that this time is set in ftp_server settimes method. Picked up from the default template.
        self.assertEqual(['-rwxrwxrwx   1 nobody   ftp            49 Jul 15 17:51 ftp_data.txt'], _pasv_list)
        # check list for active mode
        _actv_list = list()
        self.client.set_pasv(False)
        self.client.retrlines('LIST', _actv_list.append)
        # note that this time is set in ftp_server settimes method. Picked up from the default template.
        self.assertEqual(['-rwxrwxrwx   1 nobody   ftp            49 Jul 15 17:51 ftp_data.txt'], _actv_list)
        # response from active and pasv mode should be same.

    def test_nlist(self):
        # TODO: check for a user who does not have permissions to do nlst!
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        # Do a list of directory
        _pasv_list = list()
        self.client.retrlines('NLST', _pasv_list.append)
        self.assertEqual(['ftp_data.txt'], _pasv_list)
        # check list for active mode
        _actv_list = list()
        self.client.set_pasv(False)
        self.client.retrlines('NLST', _actv_list.append)
        self.assertEqual(['ftp_data.txt'], _actv_list)

    def test_retr(self):
        """Test retr or downloading a file from the server."""
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        _path = os.path.join(''.join(conpot.__path__), 'tests', 'data', 'data_temp_fs', 'ftp')
        with open(_path + '/ftp_testing_retr.txt', mode='wb') as _file:
            self.client.retrbinary("retr ftp_data.txt", _file.write)
        buffer = ''
        with open(_path + '/ftp_testing_retr.txt', mode='r') as _file:
            buffer += _file.readline()
        self.assertEqual(buffer, 'This is just a test file for Conpot\'s FTP server\n')
        os.remove(_path + '/ftp_testing_retr.txt')

    def test_rein(self):
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        self.assertEqual(self.client.sendcmd('rein'), '230 Ready for new user.')
        self.assertRaisesRegex(ftplib.error_perm, '503 Login with USER first.', self.client.sendcmd, 'pass testing')
        # TODO: Add test with existing transfer in progress.

    @freeze_time('2018-07-15 17:51:17')
    def test_stor(self):
        # let us test by uploading a file called ftp_testing.txt
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        _path = os.path.join(''.join(conpot.__path__), 'tests', 'data', 'test_data_fs', 'ftp')
        with open(_path + '/ftp_testing.txt', mode='rb') as _file:
            self.client.storbinary("stor ftp_testing_stor.txt", _file)
        self.assertIn('ftp_testing_stor.txt', self.ftp_server.handler.config.vfs.listdir('/'))
        _vfs, _data_fs = conpot_core.get_vfs('ftp')
        _vfs.remove('ftp_testing_stor.txt')
        _data_fs_file = sanitize_file_name('ftp_testing_stor.txt', self.client.sock.getsockname()[0],
                                           self.client.sock.getsockname()[1])
        _data_fs.remove(_data_fs_file)

    def test_appe(self):
        _data_1 = 'This is just a test!\n'
        _data_2 = 'This is another test\n'
        _vfs, _ = conpot_core.get_vfs('ftp')
        with _vfs.open('ftp_appe_test.txt', mode='w') as _file:
            _file.write(_data_1)
        try:
            self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
            self.client.login(user='nobody', passwd='nobody')
            _path = os.path.join(''.join(conpot.__path__), 'tests', 'data', 'data_temp_fs', 'ftp')
            with open(_path + '/ftp_appe.txt', mode='w+') as _file:
                _file.write(_data_2)
            with open(_path + '/ftp_appe.txt', mode='rb+') as _file:
                self.client.storbinary("appe ftp_appe_test.txt", _file)
            _buffer = ''
            with _vfs.open('ftp_appe_test.txt', mode='r') as _file:
                _buffer += _file.read()
            self.assertEqual(_buffer, _data_1 + _data_2)
        finally:
            _vfs.remove('ftp_appe_test.txt')

    def test_abor(self):
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        self.assertEqual(self.client.sendcmd('abor'), '225 No transfer to abort.')

    def test_rest(self):
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        # Let us test error conditions first.
        self.client.sendcmd('type i')
        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'rest')
        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'rest str')
        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'rest -1')
        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'rest 10.1')
        # REST is not supposed to be allowed in ASCII mode
        self.client.sendcmd('type a')
        self.assertRaisesRegex(ftplib.error_perm, '501 Resuming transfers not allowed in ASCII mode.',
                               self.client.sendcmd, 'rest 10')
        # Fixme: test rest while an actual transfer is going on.

    def test_stou(self):
        # fixme: incomplete test.
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.client.login(user='nobody', passwd='nobody')
        self.client.sendcmd('type i')
        self.client.sendcmd('rest 10')
        self.assertRaisesRegex(ftplib.error_temp, "Can't STOU while REST",
                               self.client.sendcmd, 'stou')

    def test_max_retries(self):
        """client should raise an error when max retries are reached."""
        self.client.connect(host='127.0.0.1', port=self.ftp_server.server.server_port)
        self.assertRaises(ftplib.error_perm, self.client.login, user='nobody', passwd='incorrect_pass')
        self.assertRaises(ftplib.error_perm, self.client.login, user='nobody', passwd='incorrect_pass')
        self.assertRaises(ftplib.error_perm, self.client.login, user='nobody', passwd='incorrect_pass')
        self.assertRaisesRegex(
            ftplib.error_temp, '421 Too many connections. Service temporarily unavailable.',
            self.client.login, user='nobody', passwd='incorrect_pass')


if __name__ == '__main__':
    unittest.main()



from struct import *
import struct
from conpot.helpers import str_to_bytes
from conpot.protocols.s7comm.exceptions import ParseException


class TPKT:
    # References: rfc2126 section-4.3, rfc1006# section-6
    # Packet format:
    # +--------+--------+----------------+-----------....---------------+
    # |version |reserved| packet length  |             TPDU             |
    # +----------------------------------------------....---------------+
    # <8 bits> <8 bits> <   16 bits    > <       variable length       >

    def __init__(self, version=3, payload=''):
        self.payload = payload
        self.version = version
        self.reserved = 0
        self.packet_length = len(payload) + 4

    def pack(self):
        return pack('!BBH', self.version, self.reserved, self.packet_length) + str_to_bytes(self.payload)

    def parse(self, packet):
        # packet = cleanse_byte_string(packet)
        try:
            # try to extract the header by pattern to find malformed header data
            header = unpack('!BBH', packet[:4])
        except struct.error:
            raise ParseException('s7comm', 'malformed packet header structure')

        # extract header data and payload
        self.version = header[0]
        self.reserved = header[1]
        self.packet_length = header[2]
        self.payload = packet[4:4 + header[2]]
        return self



# Copyright (C) 2013  Daniel creo Haslinger <creo-conpot@blackmesa.at>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# this class mediates between the SNMP attack surface and conpots databus
# furthermore it keeps request statistics iot evade being used as a DOS
# reflection tool
  
from pysnmp.smi import builder
from datetime import datetime
import conpot.core as conpot_core


class DatabusMediator(object):
    def __init__(self, oid_mappings):
        """ initiate variables """

        self.evasion_table = {}             # stores the number of requests
        self.start_time = datetime.now()
        self.oid_map = oid_mappings         # mapping between OIDs and databus keys
        self.databus = conpot_core.get_databus()

    def get_response(self, reference_class, OID):
        if OID in self.oid_map:
            if reference_class == 'DisplayString':
                (response_class,) = builder.MibBuilder().importSymbols("SNMPv2-TC", "DisplayString")

            elif reference_class == 'OctetString':
                (response_class,) = builder.MibBuilder().importSymbols("ASN1", "OctetString")

            elif reference_class == 'Integer32':
                (response_class,) = builder.MibBuilder().importSymbols("SNMPv2-SMI", "Integer32")

            elif reference_class == 'Counter32':
                (response_class,) = builder.MibBuilder().importSymbols("SNMPv2-SMI", "Counter32")

            elif reference_class == 'Gauge32':
                (response_class,) = builder.MibBuilder().importSymbols("SNMPv2-SMI", "Gauge32")

            elif reference_class == 'TimeTicks':
                (response_class,) = builder.MibBuilder().importSymbols("SNMPv2-SMI", "TimeTicks")
            # TODO: All mode classes - or autodetect'ish?
            else:
                # dynamic responses are not supported for this class (yet)
                return False
            response_value = self.databus.get_value(self.oid_map[OID])
            return response_class(response_value)
        else:
            return None

    def set_value(self, OID, value):
        # TODO: Access control. The profile shold indicate which OIDs are writable
        self.databus.set_value(self.oid_map[OID], value)

    def update_evasion_table(self, client_ip):
        """ updates dynamic evasion table """

        # get current minute as epoch..
        now = datetime.now()
        epoch_minute = int((datetime(now.year, now.month, now.day, now.hour, now.minute) -
                          datetime(1970, 1, 1)).total_seconds())

        # if this is a new minute, re-initialize the evasion table
        if epoch_minute not in self.evasion_table:
            self.evasion_table.clear()                              # purge previous entries
            self.evasion_table[epoch_minute] = {}                   # create current minute
            self.evasion_table[epoch_minute]['overall'] = 0         # prepare overall request count

        # if this is a new client, add him to the evasion table
        if client_ip[0] not in self.evasion_table[epoch_minute]:
            self.evasion_table[epoch_minute][client_ip[0]] = 0

        # increment number of requests..
        self.evasion_table[epoch_minute][client_ip[0]] += 1
        self.evasion_table[epoch_minute]['overall'] += 1

        current_numreq = self.evasion_table[epoch_minute][client_ip[0]]
        overall_numreq = self.evasion_table[epoch_minute]['overall']

        # return numreq(per_ip) and numreq(overall)
        return current_numreq, overall_numreq



# Copyright (C) 2015 Lukas Rist <glaslos@gmail.com>
#
# Rewritten by Abhinav Saxena <xandfury@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from datetime import datetime


class DotDict(dict):
    def __getattr__(self, name):
        return self[name]


class Network(object):
    def __init__(self):
        self.public_ip = None
        self.hw_address = None

    # create attributes to interface on the fly.
    def __setattr__(self, attr, value):
        object.__setattr__(self, attr, value)

    # return default value in case an attribute cannot be found in the interface
    def __getattr__(self, attr):
        raise AttributeError('Interface.Network attribute does exist')


class Interface(object):
    """ Conpot's internal interface """
    def __init__(self):
        self.network = Network()
        self.config = None
        self.protocols = DotDict()
        self.last_active = datetime.now().strftime("%b %d %Y - %H:%M:%S")

    @property
    def enabled(self):
        return [k for k in self.protocols.keys() if self.protocols[k] is not None]

    def __setattr__(self, attr, value):
        object.__setattr__(self, attr, value)

    def __getattr__(self, attr):
        raise AttributeError('Interface attribute does exist. Please check assignment')

    # FIXME: Do we really need this?
    def __repr__(self):
        s = '''          Conpot: ICS/SCADA Honeypot        
                        (c) 2018, MushMush Foundation.   
               ---------------------------------------------
                (1) Using Config                       :  {}
                (2) Enabled Protocols                  :  {}
                (3) Last Active (Attacked) on          :  {}'''.format(self.config, self.enabled, self.last_active)
        return s

    __str__ = __repr__


from datetime import datetime
import uuid

def json_default(obj):
    if isinstance(obj, datetime):
        return obj.isoformat()
    elif isinstance(obj, uuid.UUID):
        return str(obj)
    else:
        return None



# Copyright (C) 2014  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
import binascii
from conpot.helpers import chr_py3
import crc16
from . import kamstrup_constants
import conpot.core as conpot_core


logger = logging.getLogger(__name__)


class KamstrupProtocolBase(object):
    def __init__(self, communication_address):
        self.communication_address = communication_address


# ############ REQUEST MESSAGES ##############
class KamstrupRequestBase(KamstrupProtocolBase):
    def __init__(self, communication_address, command, message_bytes):
        super(KamstrupRequestBase, self).__init__(communication_address)
        self.command = command
        self.message_bytes = message_bytes
        logger.info('Kamstrup request package created with bytes: %s', self.message_bytes)

    def __str__(self):
        return 'Comm address: {0}, Command: {1}, Message: {2}'.format(hex(self.communication_address),
                                                                      hex(self.command),
                                                                      binascii.hexlify(bytearray(self.message_bytes)))


# Valid but request command unknown
class KamstrupRequestUnknown(KamstrupRequestBase):
    def __init__(self, communication_address, command_byte, message_bytes):
        super(KamstrupRequestUnknown, self).__init__(communication_address,
                                                     command_byte, message_bytes)
        logger.warning('Unknown Kamstrup request: %s', self)


class KamstrupRequestGetRegisters(KamstrupRequestBase):
    command_byte = 0x10

    def __init__(self, communication_address, command_byte, message_bytes):
        assert command_byte is command_byte
        super(KamstrupRequestGetRegisters, self).__init__(communication_address,
                                                          KamstrupRequestGetRegisters.command_byte, message_bytes)
        self.registers = []
        self._parse_register_bytes()
        logger.debug('Kamstrup request for registers: %s', str(self.registers).strip('[]'))

    def _parse_register_bytes(self):
        register_count = self.message_bytes[0]
        if len(self.message_bytes[1:]) * 2 < register_count:
            raise Exception('Invalid register count in register request')
        for count in range(register_count):
            register = self.message_bytes[1 + count * 2] * 256 + self.message_bytes[2 + count * 2]
            self.registers.append(register)


# ############ RESPONSE MESSAGES ##############
class KamstrupResponseBase(KamstrupProtocolBase):
    def __init__(self, communication_address):
        super(KamstrupResponseBase, self).__init__(communication_address)

    def serialize(self, message):
        final_message = list()

        # prefix message
        final_message.append(kamstrup_constants.RESPONSE_MAGIC)
        final_message.append(self.communication_address)

        # add the original content
        for c in message:
            final_message.append(c)

        # generate and append checksum
        crc = crc16.crc16xmodem(b''.join([chr_py3(item) for item in final_message[1:]]))
        final_message.append(crc >> 8)
        final_message.append(crc & 0xff)

        # trailing magic
        final_message.append(kamstrup_constants.EOT_MAGIC)

        escaped_message = self.escape(final_message)
        return escaped_message

    # escape everything but leading and trailing magic
    @classmethod
    def escape(cls, message):
        escaped_list = list()
        escaped_list.append(message[0])
        for c in message[1:-1]:
            if c in kamstrup_constants.NEED_ESCAPE:
                escaped_list.append(kamstrup_constants.ESCAPE)
                escaped_list.append(c ^ 0xff)
            else:
                escaped_list.append(c)
        escaped_list.append(message[-1])
        return escaped_list


class KamstrupResponseRegister(KamstrupResponseBase):
    def __init__(self, communication_address):
        super(KamstrupResponseRegister, self).__init__(communication_address)
        self.registers = []

    def add_register(self, register):
        self.registers.append(register)

    def serialize(self, message=None):
        if not message:
            message = []
        message.append(0x10)

        for register in self.registers:
            # each register must be packed: (ushort registerId, byte units, byte length, byte unknown)
            # and the following $length payload with the register value
            message.append(register.name >> 8)
            message.append(register.name & 0xff)
            message.append(register.units)
            message.append(register.length)
            # mystery byte
            message.append(register.unknown)

            low_endian_value_packed = []
            register_value = conpot_core.get_databus().get_value(register.databus_key)
            for _ in range(register.length):
                # get least significant
                low_endian_value_packed.append(register_value & 0xff)
                register_value >>= 8

            # reverse to get pack high endian
            for b in reversed(low_endian_value_packed):
                message.append(b)

        # add leading/trailing magic and escape as appropriate
        serialized_message = super(KamstrupResponseRegister, self).serialize(message)
        return bytearray(serialized_message)



# Copyright (C) 2014  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import logging
from conpot.helpers import chr_py3
import crc16
from . import kamstrup_constants


logger = logging.getLogger(__name__)


class Decoder382(object):

    REGISTERS = {

        0x01: 'Energy in',
        0x02: 'Energy out',

        0x0d: 'Energy in hi-res',
        0x0e: 'Energy out hi-res',

        0x33:  'Meter number',        # user configurable
        0x3e9: 'Meter serialnumber',  # not user configurable
        0x466: 'Meter type',
        0x417: 'Time zone',           # how is this represented?
        0x4f7: 'KMP address',
        0x4f4: 'M-bus address',

        0x041e: 'Voltage p1',
        0x041f: 'Voltage p2',
        0x0420: 'Voltage p3',

        0x0434: 'Current p1',
        0x0435: 'Current p2',
        0x0436: 'Current p3',

        0x0438: 'Power p1',
        0x0439: 'Power p2',
        0x043a: 'Power p3',

        0x178a: 'Firmware revision A',  # not too sure on how to parse there.
        0x178f: 'Firmware revision B',
    }

    def __init__(self):
        self.in_data = []
        self.in_parsing = False
        self.in_data_escaped = False
        self.out_data = []
        self.out_parsing = False
        self.out_data_escaped = False
        self.request_command_map = {0x01: self._decode_cmd_get_type,
                                    0x10: self._decode_cmd_get_register,
                                    0x92: self._decode_cmd_login}

        self.response_map = {0x10: self._decode_cmd_return_register}

    def decode_in(self, data):
        for d in data:
            d = ord(d)
            if not self.in_parsing and d != kamstrup_constants.REQUEST_MAGIC:
                logger.info('No kamstrup_meter request magic received, got: %s', d.encode('hex-codec'))
            else:
                self.in_parsing = True

                escape_escape_byte = False
                if self.in_data_escaped:
                    d ^= 0xff
                    if d is kamstrup_constants.EOT_MAGIC:
                        escape_escape_byte = True
                    self.in_data_escaped = False
                elif d is kamstrup_constants.ESCAPE:
                    self.in_data_escaped = True
                    continue

                assert self.in_data_escaped is False

                if d is kamstrup_constants.EOT_MAGIC and not escape_escape_byte:
                    if not self.valid_crc(self.in_data[1:]):
                        self.in_parsing = False
                        self.in_data = []
                        # TODO: Log discarded bytes?
                        return 'Request discarded due to invalid CRC.'
                    # now we expect (0x80, 0x3f, 0x10) =>
                    # (request magic, communication address, command byte)
                    comm_address = self.in_data[1]
                    if self.in_data[2] in self.request_command_map:
                        result = self.request_command_map[self.in_data[2]]() + ' [{0}]'.format(hex(comm_address))
                    else:
                        result = 'Unknown request command: {0}'.format(self.in_data[2])
                    self.in_data = []
                    return result
                else:
                    self.in_data.append(d)

    def decode_out(self, data):
        for d in data:
            d = ord(d)
            if not self.out_parsing and d != kamstrup_constants.RESPONSE_MAGIC:
                logger.info('Kamstrup: Expected response magic but got got: %s', d.encode('hex-codec'))
            else:
                self.out_parsing = True

                escape_escape_byte = False
                if self.out_data_escaped:
                    d ^= 0xff
                    if d is kamstrup_constants.EOT_MAGIC:
                        escape_escape_byte = True
                    self.out_data_escaped = False
                elif d is kamstrup_constants.ESCAPE:
                    self.out_data_escaped = True
                    continue

                assert self.out_data_escaped is False

                if d is kamstrup_constants.EOT_MAGIC and not escape_escape_byte:
                    if not self.valid_crc(self.out_data[1:]):
                        self.out_parsing = False
                        self.out_data = []
                        # TODO: Log discarded bytes?
                        return 'Response discarded due to invalid CRC.'
                    comm_address = self.out_data[1]
                    if self.out_data[2] in self.response_map:
                        result = self.response_map[self.out_data[2]]() + ' [{0}]'.format(hex(comm_address))
                    else:
                        result = 'Unknown response command: {0}'.format(self.out_data[2])

                    self.out_data = []
                    return result
                else:
                    self.out_data.append(d)

    def _decode_cmd_get_register(self):
        assert (self.in_data[2] == 0x10)
        cmd = self.in_data[2]
        register_count = self.in_data[3]
        message = 'Request for {0} register(s): '.format(register_count)
        if len(self.in_data[2:]) < register_count:
            return 'Invalid message, register count was too high'
        for count in range(register_count):
            register = self.in_data[4 + count * 2] * 256 + self.in_data[5 + count * 2]
            if register in Decoder382.REGISTERS:
                message += '{0} ({1})'.format(register, Decoder382.REGISTERS[register])
            else:
                message += 'Unknown ({0})'.format(register)
            if count + 1 < register_count:
                message += ', '
        return message

    def _decode_cmd_return_register(self):
        assert (self.out_data[2] == 0x10)
        # skip command bytes and trailing checksum
        msg = self.out_data[3:-2]

        return_value = 'Register reponse: '
        if len(msg) == 0:
            return_value += 'Invalid register'
        else:
            i = 0
            while i < len(msg):
                # Header is (ushort registerId, byte units, byte length, byte unknown)
                register = msg[i] * 256 + msg[i+1]
                unknown_byte_A = msg[i+2]
                length = msg[i+3]
                unknown_byte_B = msg[i+2]

                # Payload
                register_value = 0
                for p in range(length):
                    register_value += msg[i + 5 + p] << (8 * ((length - p) - 1))
                if register in Decoder382.REGISTERS:
                    return_value += '{0}({1}):{2}, '.format(register, Decoder382.REGISTERS[register], register_value)
                else:
                    return_value += '{0}:{1}, '.format(register, register_value)
                i += 5 + length

        return return_value

    # meter type
    def _decode_cmd_get_type(self):
        assert (self.in_data[2] == 0x01)
        return 'Request for GetType'

    def _decode_cmd_login(self):
        assert (self.in_data[2] == 0x92)
        pin_code = self.in_data[3] * 256 + self.in_data[4]
        return 'Login command with pin_code: {0}'.format(pin_code)

    # supplied message should be stripped of leading and trailing magic
    @classmethod
    def valid_crc(cls, message):
        supplied_crc = message[-2] * 256 + message[-1]
        calculated_crc = crc16.crc16xmodem(b''.join([chr_py3(item) for item in message[:-2]]))
        return supplied_crc == calculated_crc

    @classmethod
    def _decode_response(cls):
        return 'Decoding of this response has not been implemented yet.'



# Copyright (C) 2018  Abhinav Saxena <xandfury@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from gevent import monkey; monkey.patch_all()
import socketserver
import gevent
from gevent import queue
from gevent import select
import conpot.core as conpot_core
from conpot.core.filesystem import FilesystemError
import logging
import errno
import time
import fs
from datetime import datetime
import os
from fs import errors
from fs.path import frombase
from conpot.helpers import sanitize_file_name
from gevent import event
from conpot.protocols.ftp.ftp_utils import FTPPrivilegeException
from gevent import socket
logger = logging.getLogger(__name__)

# -----------------------------------------------------------
# Implementation Note: DTP channel that would have two queues for Input and Output (separate for producer and consumer.)
# There would be 3 threads (greenlets) in place. One for handling the command_channel, one for handling the input/
# output of the data_channel and one that would run the command processor.
# Commands class inheriting this base handler is independent of the **green drama** and may be considered on as is basis
# when migrating to async/io.
# -----------------------------------------------------------


class FTPMetrics(object):
    """Simple class to track total bytes transferred, login attempts etc."""
    def __init__(self):
        self.start_time = time.time()
        self.data_channel_bytes_recv = 0
        self.data_channel_bytes_send = 0
        self.command_chanel_bytes_send = 0
        self.command_chanel_bytes_recv = 0
        self.last_active = None
        # basically we need a timeout time composite of timeout for
        # data_sock and client_sock. Let us say that timeout of 5 sec for data_sock
        # and command_sock is 300.
        # Implementation note: Data sock is non-blocking. So the connection is closed when there is no data.

    @property
    def timeout(self):
        if self.last_active:
            return lambda: int(time.time() - self.last_active)
        else:
            return lambda: int(time.time() - self.start_time)

    def get_elapsed_time(self):
        return self.last_active - self.start_time

    def __repr__(self):
        tot = self.data_channel_bytes_recv + self.data_channel_bytes_send + self.command_chanel_bytes_recv + \
              self.command_chanel_bytes_send
        s = ''' 
        Total data transferred      : {}  (bytes)
        Command channel sent        : {}  (bytes)
        Command channel received    : {}  (bytes)
        Data channel sent           : {}  (bytes)
        Data channel received       : {}  (bytes)'''.format(tot,
                                                            self.command_chanel_bytes_send,
                                                            self.command_chanel_bytes_recv,
                                                            self.data_channel_bytes_send,
                                                            self.data_channel_bytes_recv)
        return s

    def get_metrics(self, user_name, uid, failed_login_attempts, max_login_attempts, client_address):
        s = '''
        FTP statistics for client     : {}
        ----------------------------------
        Logged in as user :{} with uid: {}
        Failed login attempts         : {}/{}
        Start time                    : {}
        Last active on                : {}
        ----------------------------------
        '''.format(client_address, user_name, uid, failed_login_attempts, max_login_attempts,
                   datetime.fromtimestamp(self.start_time).ctime(),
                   datetime.fromtimestamp(self.last_active).ctime())
        s += self.__repr__()
        return s


class FTPHandlerBase(socketserver.BaseRequestHandler):
    """Base class for a full duplex connection"""
    config = None                # Config of FTP server. FTPConfig class instance.
    host, port = None, None      # FTP Sever's host and port.
    _local_ip = '127.0.0.1'      # IP to bind the _data_listener_sock with.
    _ac_in_buffer_size = 4096    # incoming data buffer size (defaults 4096)
    _ac_out_buffer_size = 4096   # outgoing data buffer size (defaults 4096)

    def __init__(self, request, client_address, server):

        # ------------------------ Environment -------------------------
        self.client_sock = request._sock
        # only commands that are enabled should work! This is configured in the FTPConfig class.
        if not self._local_ip:
            self._local_ip = self.client_sock.getsockname()[0]  # for masquerading.. Local IP would work just fine.
        # Username of the current user
        self.username = None
        self._uid = None  # UID of the current username
        # The flag to check whether a user is already authenticated!
        self.authenticated = False

        # What commands does this FTP support
        self.COMMANDS = {}
        # conpot session
        self.session = None
        # terminate character - CR+LF
        self.terminator = b'\r\n'
        # tracking login attempts
        self.invalid_login_attempt = 0
        # max login attempts
        self.max_login_attempts = self.config.max_login_attempts

        # ftp absolute path of the file system
        self.root = self.config.vfs.getcwd()
        # get the current working directory of the current user
        self.working_dir = '/'
        # flag to check whether we need to disconnect the client
        self.disconnect_client = False
        # keep state of the last command/response
        self._last_command = None
        self._last_response = None
        # Note: From stream, block or compressed - only stream is supported.
        self._transfer_mode = None   # For APPE and REST commands
        self._restart_position = 0   # For APPE and REST commands
        # binary ('i') or ascii mode ('a')
        self._current_type = 'a'
        # buffer-size for FTP **commands**, send error if this is exceeded
        self.buffer_limit = 2048   # command channel would not accept more data than this for one command.
        self.active_passive_mode = None  # Flag to check the current mode. Would be set to 'PASV' or 'PORT'

        self._data_channel = False  # check whether the data channel is running or not. This would trigger
        # the start and end of the data channel.
        self._data_channel_send = gevent.event.Event()  # Event when we are trying to send a file.
        self._data_channel_recv = gevent.event.Event()  # Event for receiving a file.
        self.cli_ip, self.cli_port = None, None     # IP and port received from client in active/passive mode.
        self._data_sock = None
        self._data_listener_sock = None
        # socket for accepting cli_ip and cli_port in passive mode.

        self._rnfr = None  # For RNFR and RNTO
        self.metrics = FTPMetrics()  # track session related metrics.

        # Input and output queues.
        self._command_channel_input_q = gevent.queue.Queue()
        self._command_channel_output_q = gevent.queue.Queue()
        self._data_channel_output_q = gevent.queue.Queue()
        self._data_channel_input_q = gevent.queue.Queue()
        self.ftp_greenlets = None  # Keep track of all greenlets
        socketserver.BaseRequestHandler.__init__(self, request=request, client_address=client_address,
                                                 server=server)

    def ftp_path(self, path):
        """Clean and sanitize ftp paths relative fs instance it is hosted in."""
        _path = self.config.vfs.norm_path(os.path.join(self.working_dir, path))
        _path = _path.replace(self.root, '/')
        return _path

    # -- Wrappers for gevent StreamServer -------

    class false_request(object):
        def __init__(self):
            self._sock = None

        def __del__(self):
            if self._sock:
                if self._sock.fileno() != -1:
                    self._sock.close()
                del self._sock

    @classmethod
    def stream_server_handle(cls, sock, address):
        """Translate this class for use in a StreamServer"""
        request = cls.false_request()
        request._sock = sock
        server = None
        _ftp = None
        try:
            _ftp = cls(request, address, server)
        except socket.error:
            logger.warning('Unexpected Error Occurred!')
            del _ftp

    def setup(self):
        """Connect incoming connection to a FTP session"""
        self.session = conpot_core.get_session('ftp', self.client_address[0], self.client_address[1])
        logger.info('New FTP connection from {}:{}. ({})'.format(self.client_address[0], self.client_address[1],
                                                                 self.session.id))
        self.session.add_event({'type': 'NEW_CONNECTION'})
        # send 200 + banner -- new client has connected!
        self.respond(b'200 ' + self.config.banner.encode())
        #  Is there a delay in command response? < gevent.sleep(0.5) ?
        return socketserver.BaseRequestHandler.setup(self)

    def finish(self):
        """End this client session"""
        if self.disconnect_client is False:
            logger.info('FTP client {} disconnected. ({})'.format(self.client_address, self.session.id))
            self.stop_data_channel(abort=True, purge=True, reason='Closing connection to {}. '
                                                                  'Client disconnected'.format(self.client_address))
            if self._data_listener_sock:
                if self._data_listener_sock.fileno() != -1:
                    self._data_sock.close()
            socketserver.BaseRequestHandler.finish(self)
            self.client_sock.close()
            logger.info('{}'.format(self.metrics.get_metrics(
                client_address=self.client_address,
                user_name=self.username,
                uid=self._uid,
                failed_login_attempts=self.invalid_login_attempt,
                max_login_attempts=self.max_login_attempts
            )))
            self.disconnect_client = True
        else:
            logger.debug('Client {} already disconnected.'.format(self.client_address))

    def __del__(self):
        if self.disconnect_client is False:
            self.finish()
        
    # -- FTP Command Channel ------------

    def handle_cmd_channel(self):
        """Read data from the socket and add it to the _command_channel_input_q for processing"""
        log_data = dict()
        try:
            socket_read, socket_write, _ = gevent.select.select([self.client_sock], [self.client_sock], [], 1)
            # make sure the socket is ready to read - we would read from the command channel.
            if self.client_sock in socket_read:
                data = self.client_sock.recv(self.buffer_limit)
                # put the data in the _input_q for processing
                if data and data != b'':
                    log_data['request'] = data
                    if self._command_channel_input_q.qsize() > self.buffer_limit:
                        # Flush buffer if it gets too long (possible DOS condition). RFC-959 specifies that
                        # 500 response should be given in such cases.
                        logger.info('FTP command input exceeded buffer from client {}'.format(self.client_address))
                        self.respond(b'500 Command too long.')
                    else:
                        self.metrics.command_chanel_bytes_recv += len(data)
                        self.metrics.last_active = time.time()
                        self._command_channel_input_q.put(data)
            # make sure the socket is ready to write
            elif self.client_sock in socket_write and (not self._command_channel_output_q.empty()):
                response = self._command_channel_output_q.get()
                if response is not None:
                    logger.debug('Sending packet {} to client {}'.format(self.client_address, response))
                    log_data['response'] = response
                    # add len to metrics
                    self.metrics.command_chanel_bytes_send += len(response)
                    self.metrics.last_active = time.time()
                    self.client_sock.send(response)
            if 'request' in log_data or 'response' in log_data:
                logger.info('FTP traffic to {}: {} ({})'.format(self.client_address, log_data, self.session.id))
                self.session.add_event(log_data)
        except socket.error as se:
            if se.errno == errno.EWOULDBLOCK:
                gevent.sleep(0.1)
            else:
                logger.info('Socket error, remote: {}. ({}). Error {}'.format(self.client_address, self.session.id, se))
                self.session.add_event({'type': 'CONNECTION_LOST'})
                self.finish()

    def respond(self, response):
        """Send processed command/data as reply to the client"""
        response = response.encode('utf-8') if not isinstance(response, bytes) else response
        response = response + self.terminator if response[-2:] != self.terminator else response
        self._command_channel_output_q.put(response)

    def process_ftp_command(self):
        raise NotImplementedError

    # -- FTP Data Channel --------------

    def start_data_channel(self, send_recv='send'):
        """
        Starts the data channel. To be called from the command process greenlet. 
        :param send_recv: Whether the event is a send event or recv event. When set to 'send' data channel's socket
        writes data in the output queues else when set to 'read' data channel's socket reads data into the input queue.
        :type send_recv: str
        """
        try:
            assert self.cli_port and self.cli_port and self._data_sock
            if self._data_channel is True:
                logger.debug('Already sending some data that has to finish first.')
                # waait till that process finishes.
                self._data_channel_send.wait()
                self._data_channel_recv.wait()
            if send_recv is 'send':
                # we just want to do send and not receive
                self._data_channel_send.clear()
                self._data_channel_recv.set()
            else:
                # we just want to do receive and not send
                self._data_channel_recv.clear()
                self._data_channel_send.set()
            self._data_channel = True
        except AssertionError:
            self.respond(b'425 Use PORT or PASV first.')
            logger.info('Can\'t initiate {} mode since either of IP or Port supplied by the '
                        'client are None'.format(self.active_passive_mode))

    def stop_data_channel(self, abort=False, purge=False, reason=None):
        if reason:
            logger.info('Closing data channel. Reason: {}'.format(reason))
        if (not self._data_channel_output_q.empty()) or (not self._command_channel_input_q.empty()):
            if not abort:
                # Wait for all transfers to complete.
                self._data_channel_send.wait()
                self._data_channel_recv.wait()
        self._data_channel = False
        if self._data_sock and self._data_sock.fileno() != -1:
            self._data_sock.close()
        # don't want to do either send and receive.
        # Although this is done while sending - we're doing it just to be safe.
        self._data_channel_recv.set()
        self._data_channel_send.set()
        if purge:
            self.cli_ip = None
            self.cli_port = None
            # purge data in buffers .i.e the data queues.
            while self._data_channel_input_q.qsize() != 0:
                _ = self._data_channel_input_q.get()
            while self._data_channel_output_q.qsize() != 0:
                _ = self._data_channel_output_q.get()

    def handle_data_channel(self):
        if self._data_channel:
            try:
                # Need to know what kind of event are we expecting.
                if not self._data_channel_send.is_set():
                    # must be a sending event. Get from the output_q and write it to socket.
                    # pick an item from the _data_channel_output_q and send it to the requisite socket
                    if not self._data_channel_output_q.empty():
                        # Consumes data from the data channel output queue. Log it and sends it across to the client.
                        # If a file needs to be send, pass the file name directly as file parameter. sendfile is used
                        # in this case.
                        data = self._data_channel_output_q.get()
                        if data['type'] == 'raw_data':
                            logger.info('Send data {} at {}:{} for client : {}'.format(
                                data['data'], self.cli_ip, self.cli_port, self.client_address)
                            )
                            self.metrics.last_active = time.time()
                            self._data_sock.send(data=data['data'])
                            self.metrics.data_channel_bytes_send += len(data)
                        elif data['type'] == 'file':
                            file_name = data['file']
                            if self.config.vfs.isfile(file_name):
                                logger.info('Sending file {} to client {} at {}:{}'.format(
                                    file_name, self.client_address, self.cli_ip, self.cli_port)
                                )
                                try:
                                    self.metrics.last_active = time.time()
                                    with self.config.vfs.open(file_name, mode='rb') as file_:
                                        self._data_sock.sendfile(file_, 0)
                                    _size = self.config.vfs.getsize(file_name)
                                    self.metrics.data_channel_bytes_send += _size
                                except (fs.errors.FSError, FilesystemError):
                                    raise
                        if self._data_channel_output_q.qsize() == 0:
                            logger.debug('No more data to read. Either transfer finished or error occurred.')
                            self._data_channel_send.set()
                            self.respond(b'226 Transfer complete.')
                elif not self._data_channel_recv.is_set():
                    # must be a receiving event. Get data from socket and add it to input_q
                    # Receive data, log it and add it to the data channel input queue.
                    self.respond(b'125 Transfer starting.')
                    data = self._data_sock.recv(self._ac_in_buffer_size)
                    if data and data != b'':
                        self.metrics.last_active = time.time()
                        # There is some data -- could be a file.
                        logger.debug('Received {} from client {} on {}:{}'.format(
                            data, self.client_address, self.cli_ip, self.cli_port)
                        )
                        self.metrics.data_channel_bytes_recv += len(data)
                        self._data_channel_input_q.put(data)
                        while data and data != b'':
                            self.metrics.last_active = time.time()
                            data = self._data_sock.recv(self._ac_in_buffer_size)
                            logger.debug('Received {} from client {} on {}:{}'.format(
                                data, self.client_address, self.cli_ip, self.cli_port)
                            )
                            self.metrics.data_channel_bytes_recv += len(data)
                            self._data_channel_input_q.put(data)
                    # we have received all data. Time to finish this process.
                    # set the writing event to set - so that we can write this data to files.
                    self._data_channel_recv.set()
                    self.respond(b'226 Transfer complete.')
                else:
                    # assume that the read/write event has finished
                    # send a nice resp to the client saying everything has finished.
                    # set the self._data_channel(_recv/_send) markers. This would also wait for read write to finish.
                    self.stop_data_channel(reason='Transfer has completed!.')
            except (socket.error, socket.timeout) as se:
                # TODO: send appropriate response
                # Flush contents of the data channel
                reason = 'connection timed out' if isinstance(se, socket.timeout) else 'socket error'
                msg = 'Stopping FTP data channel {}:{}. Reason: {}'.format(self.cli_ip, self.cli_port, reason)
                self.stop_data_channel(abort=True, purge=True, reason=msg)
            except (fs.errors.FSError, FilesystemError, FTPPrivilegeException, FilesystemError) as fe:
                self.respond(b'550 Transfer failed.')
                self.stop_data_channel(abort=True, reason='VFS related exception occurred: {}'.format(str(fe)))

    def recv_file(self, _file, _file_pos=0, cmd='STOR'):
        """
        Receive a file - to be used with STOR, REST and APPE. A copy would be made on the _data_fs.
        :param _file: File Name to the file that would be written to fs.
        :param _file_pos: Seek file to position before receiving.
        :param cmd: Command used for receiving file.
        """
        # FIXME: acquire lock to files - both data_fs and vfs.
        with self.config.vfs.lock():
            self.start_data_channel(send_recv='recv')
            recv_err = None
            logger.info('Receiving data from {}:{}'.format(self.cli_ip, self.cli_port))
            _data_fs_file = sanitize_file_name(_file, self.client_address[0], str(self.client_address[1]))
            _data_fs_d = None
            _file_d = None
            # wait till all transfer has finished.
            self._data_channel_recv.wait()
            try:
                # we are blocking on queue for 10 seconds to wait for incoming data.
                # If there is no data in the queue. We assume that transfer has been completed.
                _data = self._data_channel_input_q.get()
                _data_fs_d = self.config.data_fs.open(path=_data_fs_file, mode='wb')
                if _file_pos == 0 and cmd == 'STOR':
                    # overwrite file or create a new one.
                    # we don't need to seek at all. Normal write process by STOR
                    _file_d = self.config.vfs.open(path=_file, mode='wb')
                else:
                    assert _file_pos != 0
                    # must seek file. This is done in append or rest(resume transfer) command.
                    # in that case, we should create a duplicate copy of this file till that seek position.
                    with self.config.vfs.open(path=_file, mode='rb') as _file_d:
                        _data_fs_d.write(_file_d.read(_file_pos))
                    # finally we should let the file to be written as requested.
                    if cmd == 'APPE':
                        _file_d = self.config.vfs.open(path=_file, mode='ab')
                    else:
                        # cmd is REST
                        _file_d = self.config.vfs.open(path=_file, mode='rb+')
                        _file_d.seek(_file_pos)
                _file_d.write(_data)
                _data_fs_d.write(_data)
                while not self._data_channel_input_q.empty():
                    _data = self._data_channel_input_q.get()
                    _file_d.write(_data)
                    _data_fs_d.write(_data)
                logger.info('Files {} and {} written successfully to disk'.format(_file, _data_fs_file))
            except (AssertionError, IOError, fs.errors.FSError, FilesystemError, FTPPrivilegeException) as fe:
                recv_err = fe
                self.stop_data_channel(abort=True, reason=str(fe))
                self.respond('554 {} command failed.'.format(cmd))
            finally:
                if _file_d and _file_d.fileno() != -1:
                    _file_d.close()
                if _data_fs_d and _data_fs_d.fileno() != -1:
                    _data_fs_d.close()
                if not recv_err:
                    self.config.vfs.chmod(_file, self.config.file_default_perms)
                    if cmd == 'STOR':
                        self.config.vfs.chown(_file, uid=self._uid, gid=self.config.get_gid(self._uid))
                    self.config.vfs.settimes(_file, accessed=datetime.now(), modified=datetime.now())
                    self.respond(b'226 Transfer complete.')

    def push_data(self, data):
        """Handy utility to push some data using the data channel"""
        # ensure data is encoded in bytes
        data = data.encode('utf8') if not isinstance(data, bytes) else data
        self._data_channel_output_q.put({'type': 'raw_data', 'data': data})

    def send_file(self, file_name):
        """Handy utility to send a file using the data channel"""
        if self._data_channel:
            self.respond("125 Data connection already open. Transfer starting.")
        else:
            self.respond("150 File status okay. About to open data connection.")
        self._data_channel_output_q.put({'type': 'file', 'file': file_name})
        self.start_data_channel()

    # -- FTP Authentication and other unities --------

    # FIXME: Refactor this. Move this to the auth module.
    def authentication_ok(self, user_pass):
        """
        Verifies authentication and sets the username of the currently connected client. Returns True or False
        Checks user names and passwords pairs. Sets the current user and uid.
        """
        # if anonymous ftp is enabled - accept any password.
        try:
            if self.username == 'anonymous' and self.config.anon_auth:
                self.authenticated = True
                self._uid = self.config.anon_uid
                self.username = self.config.user_db[self._uid]['uname']
                return True
            else:
                if (self.username, user_pass) in self.config.user_pass:
                    # user/pass match and correct!
                    self.authenticated = True
                    self._uid = self.config.get_uid(self.username)
                    return True
                return False
        except (KeyError, ValueError):
            return False
        
    # -- Actual FTP Handler -----------

    def handle(self):
        """Actual FTP service to which the user has connected."""
        while not self.disconnect_client:
            try:
                # These greenlets would be running forever. During the connection.
                # first two are for duplex command channel. Final one is for storing files on file-system.
                self.ftp_greenlets = [gevent.spawn(self.handle_cmd_channel),
                                      gevent.spawn(self.process_ftp_command),
                                      gevent.spawn(self.handle_data_channel)]
                gevent.joinall(self.ftp_greenlets)
                # Block till all jobs are not finished
            except KeyboardInterrupt:
                logger.info('Shutting FTP server.')
            finally:
                gevent.killall(self.ftp_greenlets)



# Copyright (C) 2018  Abhinav Saxena <xandfury@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


class FTPException(Exception):
    """General FTP related exceptions. """
    pass


class FTPMaxLoginAttemptsExceeded(FTPException):
    pass


class FTPPrivilegeException(FTPException):
    pass


# all commands:
ftp_commands = {
    'ABOR': dict(auth=True, perm=None, arg=False, help='Syntax: ABOR (abort transfer).'),
    'ALLO': dict(perm=None, auth=True, arg=True, help='Syntax: ALLO <SP> bytes (noop; allocate storage).'),
    'APPE': dict(perm='a', auth=True, arg=True, help='Syntax: APPE <SP> file-name (append data to file).'),
    'CDUP': dict(perm='e', auth=True, arg=False, help='Syntax: CDUP (go to parent directory).'),
    'CWD': dict(perm='e', auth=True, arg=None, help='Syntax: CWD [<SP> dir-name] (change working directory).'),
    'DELE': dict(perm='d', auth=True, arg=True, help='Syntax: DELE <SP> file-name (delete file).'),
    'HELP': dict(perm=None, auth=False, arg=None, help='Syntax: HELP [<SP> cmd] (show help).'),
    'LIST': dict(perm='l', auth=True, arg=None, help='Syntax: LIST [<SP> path] (list files).'),
    'MDTM': dict(perm='l', auth=True, arg=True, help='Syntax: MDTM [<SP> path] (file last modification time).'),
    'MODE': dict(perm=None, auth=True, arg=True, help='Syntax: MODE <SP> mode (noop; set data transfer mode).'),
    'MKD': dict(perm='m', auth=True, arg=True, help='Syntax: MKD <SP> path (create directory).'),
    'NLST': dict(perm='l', auth=True, arg=None, help='Syntax: NLST [<SP> path] (list path in a compact form).'),
    'NOOP': dict(perm=None, auth=False, arg=False, help='Syntax: NOOP (just do nothing).'),
    'PASS': dict(perm=None, auth=False, arg=None, help='Syntax: PASS [<SP> password] (set user password).'),
    'PASV': dict(perm=None, auth=True, arg=False, help='Syntax: PASV (open passive data connection).'),
    'PORT': dict(perm=None, auth=True, arg=True, help='Syntax: PORT <sp> h,h,h,h,p,p (open active data connection).'),
    'PWD': dict(perm=None, auth=True, arg=False, help='Syntax: PWD (get current working directory).'),
    'QUIT': dict(perm=None, auth=False, arg=False, help='Syntax: QUIT (quit current session).'),
    'REIN': dict(perm=None, auth=True, arg=False, help='Syntax: REIN (flush account).'),
    'RETR': dict(perm='r', auth=True, arg=True, help='Syntax: RETR <SP> file-name (retrieve a file).'),
    'RMD': dict(perm='d', auth=True, arg=True, help='Syntax: RMD <SP> dir-name (remove directory).'),
    'RNFR': dict(perm='f', auth=True, arg=True, help='Syntax: RNFR <SP> file-name (rename (source name)).'),
    'REST': dict(perm=None, auth=True, arg=True, help='Syntax: REST <SP> offset (set file offset).'),
    'RNTO': dict(perm='f', auth=True, arg=True, help='Syntax: RNTO <SP> file-name (rename (destination name)).'),
    'SITE': dict(perm=None, auth=False, arg=True, help='Syntax: SITE <SP> site-command (execute SITE command).'),
    'SITE HELP': dict(perm=None, auth=False, arg=None, help='Syntax: SITE HELP [<SP> cmd] (show SITE command help).'),
    'SITE CHMOD': dict(perm='M', auth=True, arg=True, help='Syntax: SITE CHMOD <SP> mode path (change file mode).'),
    'SIZE': dict(perm='l', auth=True, arg=True, help='Syntax: SIZE <SP> file-name (get file size).'),
    'STAT': dict(perm='l', auth=False, arg=None, help='Syntax: STAT [<SP> path name] (server stats [list files]).'),
    'STOR': dict(perm='w', auth=True, arg=True, help='Syntax: STOR <SP> file-name (store a file).'),
    'STOU': dict(perm='w', auth=True, arg=None, help='Syntax: STOU [<SP> name] (store a file with a unique name).'),
    'STRU': dict(perm=None, auth=True, arg=True, help='Syntax: STRU <SP> type (noop; set file structure).'),
    'SYST': dict(perm=None, auth=False, arg=False, help='Syntax: SYST (get operating system type).'),
    'TYPE': dict(perm=None, auth=True, arg=True, help='Syntax: TYPE <SP> [A | I] (set transfer type).'),
    'USER': dict(perm=None, auth=False, arg=True, help='Syntax: USER <SP> user-name (set username).'),
    'XCUP': dict(perm='e', auth=True, arg=False, help='Syntax: XCUP (obsolete; go to parent directory).'),
    'XCWD': dict(perm='e', auth=True, arg=None, help='Syntax: XCWD [<SP> dir-name] (obsolete; change directory).'),
    'XMKD': dict(perm='m', auth=True, arg=True, help='Syntax: XMKD <SP> dir-name (obsolete; create directory).'),
    'XPWD': dict(perm=None, auth=True, arg=False, help='Syntax: XPWD (obsolete; get current dir).'),
    'XRMD': dict(perm='d', auth=True, arg=True, help='Syntax: XRMD <SP> dir-name (obsolete; remove directory).')
}


def get_data_from_iter(iterator):
    """This utility function generates data from iterators and returns them as string"""
    buffer = ''
    try:
        while True:
            buffer += str(next(iterator))
    except StopIteration:
        return buffer


# Copyright (C) 2014  Lukas Rist <glaslos@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import gevent.monkey; gevent.monkey.patch_all()
import unittest
import conpot.utils.ext_ip
from gevent.pywsgi import WSGIServer
import gevent


class TestExtIPUtil(unittest.TestCase):

    def setUp(self):
        def application(environ, start_response):
            headers = [('Content-Type', 'text/html')]
            start_response('200 OK', headers)
            return [b'127.0.0.1']

        self.server = WSGIServer(('localhost', 8000), application)
        gevent.spawn(self.server.serve_forever)

    def tearDown(self):
        self.server.stop()

    def test_ip_verify(self):
        self.assertTrue(conpot.utils.ext_ip._verify_address("127.0.0.1") is True)

    def test_ext_util(self):
        ip_address = str(conpot.utils.ext_ip._fetch_data(urls=["http://127.0.0.1:8000", ]))
        self.assertTrue(conpot.utils.ext_ip._verify_address(ip_address) is True)

    def test_fetch_ext_ip(self):
        self.assertIsNotNone(conpot.utils.ext_ip.get_ext_ip(urls=["https://api.ipify.org"]))


if __name__ == '__main__':
    unittest.main()



# Copyright (C) 2014  Johnny Vestergaard <jkv@unixcluster.dk>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import socket
import binascii
import random
import conpot
from gevent.server import StreamServer
import gevent
from conpot.helpers import chr_py3
import conpot.core as conpot_core
from conpot.protocols.kamstrup.meter_protocol import request_parser
from conpot.protocols.kamstrup.meter_protocol.command_responder import CommandResponder
from conpot.core.protocol_wrapper import conpot_protocol
# import logging as logger
# import sys
# logger.basicConfig(stream=sys.stdout, level=logger.DEBUG)
import logging
logger = logging.getLogger(__name__)


@conpot_protocol
class KamstrupServer(object):
    def __init__(self, template, template_directory, args, timeout=0):
        self.timeout = timeout
        self.command_responder = CommandResponder(template)
        self.server_active = True
        self.server = None
        conpot_core.get_databus().observe_value('reboot_signal', self.reboot)
        logger.info('Kamstrup protocol server initialized.')

    # pretending reboot... really just closing connecting while "rebooting"
    def reboot(self, key):
        assert (key == 'reboot_signal')
        self.server_active = False
        logger.info('Pretending server reboot')
        gevent.spawn_later(2, self.set_reboot_done)

    def set_reboot_done(self):
        logger.info('Stopped pretending reboot')
        self.server_active = True

    def handle(self, sock, address):
        session = conpot_core.get_session('kamstrup_protocol', address[0], address[1])
        logger.info('New Kamstrup connection from %s:%s. (%s)', address[0], address[1], session.id)
        session.add_event({'type': 'NEW_CONNECTION'})

        self.server_active = True

        parser = request_parser.KamstrupRequestParser()
        try:
            while self.server_active:
                raw_request = sock.recv(1024)

                if not raw_request:
                    logger.info('Kamstrup client disconnected. (%s)', session.id)
                    session.add_event({'type': 'CONNECTION_LOST'})
                    break

                for x in raw_request:
                    parser.add_byte(chr_py3(x))

                while True:
                    request = parser.get_request()
                    if not request:
                        session.add_event({'type': 'CONNECTION_LOST'})
                        break
                    else:
                        logdata = {'request': binascii.hexlify(bytearray(request.message_bytes))}
                        response = self.command_responder.respond(request)
                        # real Kamstrup meters has delay in this interval
                        gevent.sleep(random.uniform(0.24, 0.34))
                        if response:
                            serialized_response = response.serialize()
                            logdata['response'] = binascii.hexlify(serialized_response)
                            logger.info('Kamstrup traffic from %s: %s (%s)', address[0], logdata, session.id)
                            sock.send(serialized_response)
                            session.add_event(logdata)
                        else:
                            session.add_event(logdata)
                            break

        except socket.timeout:
            logger.debug('Socket timeout, remote: %s. (%s)', address[0], session.id)
            session.add_event({'type': 'CONNECTION_LOST'})

        sock.close()

    def start(self, host, port):
        connection = (host, port)
        self.server = StreamServer(connection, self.handle)
        logger.info('Kamstrup protocol server started on: %s', connection)
        self.server.serve_forever()

    def stop(self):
        self.server.stop()


if __name__ == '__main__':
    TCP_IP = '127.0.0.1'
    TCP_PORT = 1025
    import os
    dir_name = os.path.dirname(conpot.__file__)
    conpot_core.get_databus().initialize(dir_name + '/templates/kamstrup_382/template.xml')
    server = KamstrupServer(dir_name + '/templates/kamstrup_382/kamstrup_meter/kamstrup_meter.xml', None, None)
    try:
        server.start(TCP_IP, TCP_PORT)
    except KeyboardInterrupt:
        server.stop()